#!/usr/bin/env python3
# -*- Coding: UTF-8 -*-

# ---------------------------------------------------------------------------
# Open Asset Import Library (ASSIMP)
# ---------------------------------------------------------------------------
#
# Copyright (c) 2006-2010, ASSIMP Development Team
#
# All rights reserved.
#
# Redistribution and use of this software in source and binary forms, 
# with or without modification, are permitted provided that the following 
# conditions are met:
# 
# * Redistributions of source code must retain the above
#   copyright notice, this list of conditions and the
#   following disclaimer.
# 
# * Redistributions in binary form must reproduce the above
#   copyright notice, this list of conditions and the
#   following disclaimer in the documentation and/or other
#   materials provided with the distribution.
# 
# * Neither the name of the ASSIMP team, nor the names of its
#   contributors may be used to endorse or promote products
#   derived from this software without specific prior
#   written permission of the ASSIMP Development Team.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# ---------------------------------------------------------------------------

"""Shared stuff for the gen_db and run scripts """

# -------------------------------------------------------------------------------
def hashing(file,pp):
    """ Map an input file and a postprocessing config to an unique hash.

    The hash is used to store the item in the database. It 
    needs to be persistent across different python implementations
    and platforms, so we implement the hashing manually.
    """

    def myhash(instring):
        # sdbm hash
        res = 0
        for t in instring:
            res = (ord(t) + (res<<6) + (res<<16) - res) % 2**32
        return res

    return hex(myhash(file.replace('\\','/')+":"+pp))


assimp_bin_path = None
# -------------------------------------------------------------------------------
def find_assimp_or_die():
    """Find assimp_cmd's binary for the current platform.
    
    The path to the binary is stored in assimp_bin_path, the process
    is aborted if it can't be found.

    """

    import os
    import platform
    import sys

    def locate_file(f_list):
        for f in f_list:
            try:
                fl = open(f,"rb")
            except IOError:
                continue
            fl.close()
            return f
        return None

    global assimp_bin_path
    if os.name == "nt":
        search_x86 = [
            os.path.join("..","..","bin","assimpcmd_release-dll_Win32","assimp.exe"),
            os.path.join("..","..","bin","x86","assimp"),
            os.path.join("..","..","bin","Release","assimp.exe")
        ]
        if platform.machine() == "x86":
            search = search_x86
        else: # amd64, hopefully
            search = [
                os.path.join("..","..","bin","assimpcmd_release-dll_x64","assimp.exe"),
                os.path.join("..","..","bin","x64","assimp")
            ]
            # x64 platform does not guarantee a x64 build. Also look for x86 as last paths.
            search += search_x86
        
        assimp_bin_path = locate_file(search)
        if assimp_bin_path is None:
            print("Can't locate assimp_cmd binary")
            print("Looked in", search)
            sys.exit(-5)

        print("Located assimp/assimp_cmd binary from", assimp_bin_path)
    elif os.name == "posix":
        #search = [os.path.join("..","..","bin","gcc","assimp"),
        #    os.path.join("/usr","local","bin",'assimp')]
        assimp_bin_path = "assimp"
        print("Taking system-wide assimp binary")
    else:
        print("Unsupported operating system")
        sys.exit(-5)

if __name__ == '__main__':
    find_assimp_or_die()

 # vim: ai ts=4 sts=4 et sw=4



# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2005
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file base.py Widget Base Class

import sys, wx
import host, events, language, paths
import sb.confdb as st
import widgets


# Font styles.
NORMAL = 0
BOLD = 1
ITALIC = 2
HEADING = 3
TITLE = 4
SMALL = 5


class Widget:
    """Widget is the abstract base class for all user interface
    widgets.  Each subclass acts as a wrapper around the real
    wxWidget, converting wxWidgets events to commands and
    notifications."""

    def __init__(self, wxWidget):
        """Constructor.
        @param wxWidget The wxWidget this Widget manages.
        """
        self.wxWidget = wxWidget
        self.menuItems = None
        self.focusId = ''
        self.menuId = None

        # Callback functions for listening directly to changes.
        self.reactors = []

        self.setNormalStyle()

        # Listen for right mouse clicks in case a popup menu should be
        # displayed.
        wx.EVT_RIGHT_DOWN(wxWidget, self.onRightClick)
        wx.EVT_RIGHT_UP(wxWidget, self.onRightClick)

        # A notification is sent when the widget is entered.
        wx.EVT_ENTER_WINDOW(wxWidget, self.onEnterWidget)

        # Listen for focus changes.
        wx.EVT_SET_FOCUS(wxWidget, self.onFocus)

        # Register a notification listener. These notifications are expected
        # to be handled by all widgets.
        events.addNotifyListener(self.onNotify, ['language-changed',
                                                 'focus-request'])
        # Widgets are not automatically listening to commands.
        #events.addCommandListener(self.onCommand)

    def getWxWidget(self):
        """Returns the real wxWidget of this widget."""
        return self.wxWidget

    def addProfileChangeListener(self):
        """Adds a listener callback that listens to changes of the active
        profile. Only those widgets that manage a setting should need this."""
        
        events.addNotifyListener(self.onNotify, ['active-profile-changed'])                                 

    def addValueChangeListener(self):
        """Adds a listener callback that listens to value change notifications
        sent with this widget's identifier."""

        events.addNotifyListener(self.onNotify, 
                                 [self.widgetId + '-value-changed'])

    def clear(self):
        """This is called when the widget is removed from the area by
        calling Area.clear().  This is not called when the area gets
        destroyed."""
        pass

    def destroy(self):
        """Destroy the widget."""

        # Remove the notification listener of this widget.
        events.removeNotifyListener(self.onNotify)
        events.removeCommandListener(self.onCommand)

        w = self.getWxWidget()

        # Detach from the containing sizer.
        sizer = w.GetContainingSizer()
        if sizer:
            sizer.Detach(w)

        w.Destroy()

    def retranslate(self):
        """Update the label(s) of the widget by retranslating them.
        This is called when the language string database changes."""
        pass

    def setFocusId(self, identifier):
        """Setting the focus identifier causes the widget to send
        focus notifications using the identifier.

        @param identifier A string.
        """
        self.focusId = identifier

    def focus(self):
        try:
            self.getWxWidget().SetFocus()
        except:
            pass

    def enable(self, doEnable=True):
        """Enable or disable the widget for user input.

        @param doEnable Boolean value.  True if the widget should be
        enabled, fFalse if not.
        """
        self.getWxWidget().Enable(doEnable)

    def disable(self):
        """Update the label(s) of the widget by retranslating them.
        This is called when the language string database changes."""
        self.enable(False)

    def setMinSize(self, width, height):
        """Set a minimum size for the widget."""
        self.getWxWidget().SetMinSize((width, height))

    def setPopupMenuId(self, menuId):
        """Define a popup menu identifier. When the identifier has been defined, 
        a notification is sent just prior to showing a popup menu, allowing 
        adapting the contents of the menu to the current situation."""
        
        self.menuId = menuId

    def setPopupMenu(self, menuItems):
        """Set a popup menu that is displayed when the widget is
        right-clicked (Ctrl-Clicked).

        @param menuItems An array of identifiers.  These will be sent
        as Command events when items are clicked.  Set to None if the
        menu should be disabled.  The array may also contain tuples:
        (identifier, command).
        """
        self.menuItems = menuItems

    def onRightClick(self, ev):
        # Let wxWidgets handle the event, too.
        widget = self.getWxWidget()
        if not widget.IsEnabled():
            return
            
        # Going up or down?
        if ev.ButtonDown():
            # Mouse right down doesn't do anything.
            return

        # Request an update to the popup menu.
        if self.menuId:
            events.send(events.Notify(self.menuId + '-update-request'))

        # Display the popup menu.
        if self.menuItems:
            menu = wx.Menu()
            self.commandMap = {}

            idCounter = 10000

            # Create the menu items.
            for item in self.menuItems:
                if type(item) == tuple:
                    itemId = item[0]
                    itemCommand = item[1]
                else:
                    itemId = item
                    itemCommand = item

                if itemId == '-':
                    # This is just a separator.
                    menu.AppendSeparator()
                    continue
                
                # Generate a new ID for the item.
                wxId = idCounter
                idCounter += 1
                self.commandMap[wxId] = itemCommand
                menu.Append(wxId,
                            widgets.uniConv(language.translate('menu-' + itemId)))
                wx.EVT_MENU(widget, wxId, self.onPopupCommand)

            # Display the menu.  The callback gets called during this.
            widget.PopupMenu(menu, ev.GetPosition())
            menu.Destroy()            

    def onPopupCommand(self, ev):
        events.send(events.Command(self.commandMap[ev.GetId()]))

    def onFocus(self, event):
        """Handle the wxWidgets event that's sent when the widget
        received input focus.

        @param event A wxWidgets event.
        """
        # Send a focus event if a focus identifier has been specified.
        if self.focusId:
            events.send(events.FocusNotify(self.focusId))
        event.Skip()

    def onEnterWidget(self, event):
        """Handle the wxWidget event of entering the window."""

        if self.focusId:
            events.send(events.FocusNotify(self.focusId))
        event.Skip()

    def onNotify(self, event):
        """Handle a notification.  All widgets are automatically
        registered for listening to notifications.

        @param event An event.Notify object.
        """
        # Has someone requested this widget to receive focus?
        if (event.hasId('focus-request') and 
            event.getFocus() == self.focusId):
            self.focus()

        elif event.hasId('language-changed'):
            self.retranslate()
        
    def onCommand(self, event):
        """Handle a command.  All widgets are automatically registered
        for listening to commands.

        @param event An event.Command object.
        """
        # By default we do nothing.
        pass

    def addReaction(self, func):
        """Add a new callback function that will be called at suitable
        times when a reaction is needed (depends on widget type).

        @param func Callback function.
        """

        self.reactors.append(func)

    def react(self):
        """Call all the reaction callback functions."""

        for func in self.reactors:
            func()

    def setStyle(self, textStyle):
        """Set a new style for the text label.
        @param style One of the constants:
        - Widget.NORMAL: default style
        - Widget.BOLD: bold typeface
        - Widget.ITALIC: italic typeface
        - Widget.HEADING: large bold typeface
        - Widget.TITLE: largest bold typeface
        """
        weight = wx.NORMAL
        style = wx.NORMAL
        styleName = 'style-' + \
                    ['normal', 'bold', 'italic', 'heading', 'title', 
                     'small'][textStyle] + '-'

        try:
            size = st.getSystemInteger(styleName + 'size')
        except:
            size = 11

        try:
            font = st.getSystemString(styleName + 'font')
        except:
            font = ''

        try:
            w = st.getSystemString(styleName + 'weight')
            if w == 'bold':
                weight = wx.BOLD
        except:
            pass

        try:
            s = st.getSystemString(styleName + 'slant')
            if s == 'italic':
                style = wx.ITALIC
        except:
            pass

        newFont = wx.Font(size, wx.DEFAULT, style, weight, faceName=font)
        w = self.getWxWidget()
        w.SetFont(newFont)

        #w.SetSize(w.GetBestSize())
        #bestSize = w.GetBestSize()
        #w.SetMinSize((0,bestSize[1]))

        self.updateDefaultSize()

    def updateDefaultSize(self):
        """By default widgets can be scaled horizontally to zero width."""
        w = self.getWxWidget()
        bestSize = w.GetBestSize()
        w.SetMinSize((0, bestSize[1]))

    def resizeToBestSize(self):
        w = self.getWxWidget()
        bestSize = w.GetBestSize()
        w.SetMinSize(bestSize)
        w.SetSize(bestSize)      
        
    def setNormalStyle(self):
        """Set the style of the text label to Widget.NORMAL."""
        self.setStyle(NORMAL)

    def setSmallStyle(self):
        """Set the style of the text label to Widget.SMALL."""
        self.setStyle(SMALL)

    def setBoldStyle(self):
        """Set the style of the text label to Widget.BOLD."""
        self.setStyle(BOLD)

    def setItalicStyle(self):
        """Set the style of the text label to Widget.ITALIC."""
        self.setStyle(ITALIC)

    def setHeadingStyle(self):
        """Set the style of the text label to Widget.HEADING."""
        self.setStyle(HEADING)

    def setTitleStyle(self):
        """Set the style of the text label to Widget.TITLE."""
        self.setStyle(TITLE)

    def freeze(self):
        """Prevent updates from occurring on the screen."""
        self.getWxWidget().Freeze()

    def unfreeze(self):
        """Allow updates to occur."""
        self.getWxWidget().Thaw()
        self.getWxWidget().Refresh()
        



#!/usr/bin/env python
# Assumes qhelpgenerator is on the path (part of Qt)
import re, sys, os, string

print 'Patching index.qhp...'
input = file(sys.argv[1], 'rt').read()

# Doxygen writes a spurious </section> tag for some reason.
#output = re.sub('</section>\s+<section title="Namespace Members"', '<!-- patched -->\n<section title="Namespace Members"', input)
output = re.sub('ref="index.html#Overview" />\n\s+</section>', 'ref="index.html#Overview" />\n<!-- patched -->', input)
output = re.sub('</section>\s+</section>\s+<section title="Math Utilities"', 
    '<!--patched--></section>\n<section title="Math Utilities"', output)
#output = re.sub('</section>\s+</toc>', '\n<!-- patched --></toc>', output)

# Rewrite the source file.
file(sys.argv[1], 'wt').write(output)

os.system('qhelpgenerator ' + string.join(sys.argv[1:], ' '))



# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2005
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file text.py Text Widgets

import sys
import wx
import wx.html
import wx.lib.fancytext as fancy
import host, events, language, paths
import sb.confdb as st
import base
from widgets import uniConv

# Workaround for a wxWidgets StaticText bug.
if host.isMac():
    from wx.lib.stattext import GenStaticText as StaticText
else:
    from wx import StaticText

def breakLongLines(text, maxLineLength):
    """Break long lines with newline characters.

    @param maxLineLength  Maximum length of a line, in characters.

    @return The new text with newlines inserted."""
    
    brokenText = ''
    lineLen = 0
    skipping = False
    for i in range(len(text) - 1):
        if text[i] == '<':
            skipping = True
        elif skipping and text[i] == '>':
            skipping = False

        if not skipping:
            lineLen += 1
                    
        if (text[i] in ' -/:;') and lineLen > maxLineLength:
            # Break it up!
            if text[i] != ' ':
                brokenText += text[i]
            if not skipping:
                # Newlines are not inserted if we're skipping.
                brokenText += "\n"
                lineLen = 0
        else:
            brokenText += text[i]

    return brokenText + text[-1]   


class MyStaticText (StaticText):
    def __init__(self, parent, wxId, label, style=0):
        StaticText.__init__(self, parent, wxId, label, style=style)
        if host.isWindows():
            # No background.
            self.SetBackgroundStyle(wx.BG_STYLE_SYSTEM)
        wx.EVT_ERASE_BACKGROUND(self, self.clearBack)

    def clearBack(self, event):
        pass


class Text (base.Widget):
    """A static text widget that displays a string of text."""

    # Alignments.
    LEFT = 0
    MIDDLE = 1
    RIGHT = 2
    
    def __init__(self, parent, wxId, id, suffix='', maxLineLength=0, align=0):
        """Constructor.

        @param parent
        @param wxId
        @param id      Identifier of the text string.
        @param suffix  Suffix for the translated version of the text.
        @param maxLineLength  Maximum line length for the text (in characters).
        @param alignment      Alignment (LEFT, MIDDLE, RIGHT).
        """
        # Prepare the text string.
        self.maxLineLength = maxLineLength
        self.suffix = suffix
        self.widgetId = id

        # Alignment style flag.
        styleFlags = 0
        if align == Text.LEFT:
            styleFlags |= wx.ALIGN_LEFT
        elif align == Text.MIDDLE:
            styleFlags |= wx.ALIGN_CENTRE
        elif align == Text.RIGHT:
            styleFlags |= wx.ALIGN_RIGHT
            
        base.Widget.__init__(self, MyStaticText(parent, wxId,
                                                uniConv(self.__prepareText()),
                                                style = styleFlags |
                                                wx.ST_NO_AUTORESIZE))
        self.setNormalStyle()

        # When the text is clicked, send a notification.
        wx.EVT_LEFT_DOWN(self.getWxWidget(), self.onClick)

    def __prepareText(self):
        text = language.translate(self.widgetId) + self.suffix
        if self.maxLineLength > 0:
            text = breakLongLines(text, self.maxLineLength)
        return text

    def retranslate(self):
        if self.widgetId:
            self.setText(self.__prepareText())

    def setText(self, text):
        "Set new untranslated content into the text widget."
        self.getWxWidget().SetLabel(uniConv(text))

    def onClick(self, ev):
        """Clicking a label causes a focus request to be sent."""
        
        focusId = self.focusId
        if not focusId:
            focusId = self.widgetId
        
        if focusId:
            event = events.FocusRequestNotify(focusId)
            events.send(event)

        ev.Skip()


class MyHtmlWindow (wx.html.HtmlWindow):
    def __init__(self, parent, wxId):
        wx.html.HtmlWindow.__init__(self, parent, wxId)
        #self.SetBackgroundStyle(wx.BG_STYLE_COLOUR)
        #self.SetBackgroundColour(wx.SystemSettings_GetColour(
        #    wx.SYS_COLOUR_WINDOW))
        self.SetBorders(3)

    def OnLinkClicked(self, link):
        """Handle a hypertext link click by sending a command event.

        @param link A wxHtmlLinkInfo object.
        """
        events.send(events.Command(link.GetHref()))


class FormattedText (base.Widget):
    def __init__(self, parent, wxId, useHtmlFormatting=True, initialText=''):
        """Constructor.

        @param parent  Parent wxWidget ID.
        @param wxId    ID of the formatted text widget.
        @param useHtmlFormatting  Use a HTML window widget.
        @param initialText  Initial contents.
        """

        self.useHtml = useHtmlFormatting
        
        if self.useHtml:
            base.Widget.__init__(self, MyHtmlWindow(parent, wxId))
        else:
            if host.isMac():
                # It appears wxWidgets fancy text is broken on the Mac.
                useFancy = True #False
            else:
                useFancy = True
            
            if useFancy:
                text = initialText.replace('<b>', '<font weight="bold">')
                text = text.replace('</b>', '</font>')
                text = text.replace('<i>', '<font style="italic">')
                text = text.replace('</i>', '</font>')
                text = text.replace('<tt>', '<font family="fixed">')
                text = text.replace('</tt>', '</font>')

                text = '<font family="swiss" size="%s">' \
                       % st.getSystemString('style-normal-size') + text + '</font>'

                # fancytext doesn't support non-ascii chars?
                text = text.replace('ä', 'a')
                text = text.replace('ö', 'o')
                text = text.replace('ü', 'u')
                text = text.replace('å', 'a')
                text = text.replace('Ä', 'A')
                text = text.replace('Ö', 'O')
                text = text.replace('Ü', 'U')
                text = text.replace('Å', 'A')

            else:
                text = initialText.replace('<b>', '')
                text = text.replace('</b>', '')
                text = text.replace('<i>', '')
                text = text.replace('</i>', '')
                text = text.replace('<tt>', '')
                text = text.replace('</tt>', '')

            # Break it up if too long lines detected.
            brokenText = breakLongLines(text, 70)
            
            if useFancy:
                base.Widget.__init__(self, fancy.StaticFancyText(
                    parent, wxId, uniConv(brokenText)))
            else:
                base.Widget.__init__(self, StaticText(parent, wxId, 
                    uniConv(brokenText)))

            self.resizeToBestSize()

    def setText(self, formattedText):
        """Set new HTML content into the formatted text widget."""
        w = self.getWxWidget()

        if self.useHtml:
            fontName = None
            fontSize = None
            
            if st.isDefined("style-html-font"):
                fontName = st.getSystemString("style-html-font")
            if st.isDefined("style-html-size"):
                fontSize = st.getSystemString("style-html-size")
            
            if fontName != None or fontSize != None:
                fontElem = "<font"
                if fontName != None:
                    fontElem += ' face="%s"' % fontName
                if fontSize != None:
                    fontElem += ' size="%s"' % fontSize
                fontElem += ">"
            else:
                fontElem = None

            #color = wx.SystemSettings_GetColour(wx.SYS_COLOUR_HIGHLIGHT)
            #print color
        
            #formattedText = '<body bgcolor="#%02x%02x%02x">' \
            #                % (color.Red(), color.Green(), color.Blue()) \
            #                + formattedText + '</body>'
        
            if fontElem == None:
                w.SetPage(uniConv(formattedText))
            else:
                w.SetPage(uniConv(fontElem + formattedText + '</font>'))

    def setSystemBackground(self):
        """Use the window background colour as the background colour
        of the formatted text."""

        self.getWxWidget().SetBackgroundStyle(wx.BG_STYLE_SYSTEM)

        #color = wx.SystemSettings_GetColour(wx.SYS_COLOUR_WINDOW)
        #print color



#!/usr/bin/env python2.7
#
# Command line utility for zipping a directory of files into a .pack.
# Checks that the required Info file is present.  
#
# Usage:
#   buildpackage (pack-dir) (output-dir)
#

import sys, os, os.path, zipfile

if len(sys.argv) < 2:
    print "Usage: %s (pack-dir) (output-dir)" % sys.argv[0]
    sys.exit(0)

# Check quiet flag.
class Package:
    def __init__(self, sourcePath):
        self.sourcePath = sourcePath
        
    def build(self, outputPath):
        # Ensure the output path exists.
        try:
            os.makedirs(outputPath)
        except:
            pass
    
        outputName = os.path.join(outputPath, os.path.basename(self.sourcePath))
        pack = zipfile.ZipFile(outputName, 'w', zipfile.ZIP_DEFLATED)        
        contents = []
            
        # Index the contents of the folder recursively.
        def descend(path):
            for name in os.listdir(os.path.join(self.sourcePath, path)):
                if name[0] == '.':
                    continue # Ignore these.

                if len(path):
                    internalPath = os.path.join(path, name)
                else:
                    internalPath = name
                fullPath = os.path.join(self.sourcePath, internalPath)
                    
                if os.path.isfile(fullPath):
                    contents.append((fullPath, internalPath))
                elif os.path.isdir(fullPath):
                    descend(internalPath)
        descend('')

        # Check for the required metadata file.
        foundInfo = False
        for full, internal in contents:
            if internal.lower() == 'info' or internal.lower() == 'info.dei':
                foundInfo = True
                break            
        if not foundInfo:
            print "No 'Info' file found in \"%s\"!" % self.sourcePath
            sys.exit(1)
        
        # Write entries in alphabetical order.
        for full, internal in sorted(contents):
            pack.write(full, internal)
            
        # Write it out.
        print "Wrote %s (contains %i files)." % (outputName, len(pack.namelist()))
        pack.close()

if __name__ == "__main__":
    p = Package(sys.argv[1])
    p.build(sys.argv[2])



# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2006
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file list.py List Widgets

import wx
import host, events, language, paths
import sb.profdb as pr
import sb.confdb
from widgets import uniConv
import base


class List (base.Widget):
    """List control widget.  Supports several styles."""

    # List styles.
    STYLE_SINGLE = 0
    STYLE_MULTIPLE = 1
    STYLE_CHECKBOX = 2
    STYLE_COLUMNS = 3
    
    def __init__(self, parent, wxId, id, style=0):
        self.style = style
        
        if style == List.STYLE_CHECKBOX:
            # Listbox with checkable items.
            base.Widget.__init__(self, wx.CheckListBox(parent, wxId))
        elif style == List.STYLE_COLUMNS:
            base.Widget.__init__(self, wx.ListCtrl(parent, wxId,
                                                   style=wx.LC_REPORT))
                                                          #wx.SIMPLE_BORDER)))
            wx.EVT_LEFT_DOWN(self.getWxWidget(), self.onLeftUp)
            wx.EVT_LIST_ITEM_SELECTED(parent, wxId, self.onItemSelected)
            wx.EVT_LIST_ITEM_DESELECTED(parent, wxId, self.onItemDeselected)
            wx.EVT_LIST_COL_CLICK(parent, wxId, self.onColumnClick)
        else:
            # Normal listbox.
            if style == List.STYLE_MULTIPLE:
                styleFlags = wx.LB_EXTENDED
            elif style == List.STYLE_SINGLE:
                styleFlags = wx.LB_SINGLE
            base.Widget.__init__(self, wx.ListBox(parent, wxId, style=styleFlags))

        # Will be used when sending notifications.
        self.widgetId = id

        self.expandingColumnIndex = -1

        self.items = []
        self.columns = []
        
        wx.EVT_SIZE(self.getWxWidget(), self.onSize)

    def setImageList(self, imageList):
        self.getWxWidget().SetImageList(imageList, wx.IMAGE_LIST_SMALL)

    def onLeftUp(self, ev):
        """Handle the wxWidgets event when the user clicks on the list widget.

        @param ev A wxWidgets event object.
        """
        w = self.getWxWidget()

        item, flags = w.HitTest(ev.GetPosition())
        if item < 0:
            ev.Skip()
            return

        if flags & wx.LIST_HITTEST_ONITEMICON:
            e = events.Notify(self.widgetId + '-icon-click')
            e.item = self.items[w.GetItemData(item)]
            events.send(e)
           
        ev.Skip()
        
    def onColumnClick(self, ev):
        events.send(events.Notify(self.columns[ev.GetColumn()] + '-column-click'))
        ev.Skip()
        
    def onSize(self, event):
        # Resize the expanding column.
        if self.expandingColumnIndex >= 0:
            # This is the new width.
            widget = self.getWxWidget()
            width = widget.GetClientSizeTuple()[0]
            # How much left from the other columns?
            total = 0
            for i in range(widget.GetColumnCount()):
                if i != self.expandingColumnIndex: 
                    total += widget.GetColumnWidth(i)
            widget.SetColumnWidth(self.expandingColumnIndex, width - total - 
                wx.SystemSettings_GetMetric(wx.SYS_VSCROLL_X))

        event.Skip()    

    def addItem(self, identifier, isChecked=False):
        """Append a new item into the listbox.

        @param identifier Identifier of the item.
        """
        w = self.getWxWidget()
        if language.isDefined(identifier):
            visibleText = language.translate(identifier)
        else:
            visibleText = identifier
        w.Append(uniConv(visibleText))
        self.items.append(identifier)

        # In a checklistbox, the items can be checked.
        if self.style == List.STYLE_CHECKBOX:
            if isChecked:
                w.Check(w.GetCount() - 1)

    def removeItem(self, identifier):
        """Remove an item from the listbox.

        @param identifier  Identifier of the item to remove.
        """
        w = self.getWxWidget()
        for i in range(len(self.items)):
            if identifier == self.items[i]:
                w.Delete(i)
                self.items.remove(identifier)
                break

    def addColumn(self, identifier, width=None):
        """Append a column into a list that uses the column style.
        
        @param identifier  Column identifier.
        @param width  Width of the column in pixels. One of the columns
                      can use None for width, which means it is automatically
                      expanded to fill all available space.
        """
        w = self.getWxWidget()
        numCols = w.GetColumnCount()
        if width is None:
            colWidth = 1
            # This column will expand to fill all remaining space.
            self.expandingColumnIndex = len(self.columns)
        else:
            colWidth = width
            
        self.getWxWidget().InsertColumn(numCols,
                                        language.translate(identifier),
                                        width=colWidth)
        self.columns.append(identifier)
        
    def retranslateColumns(self):
        w = self.getWxWidget()
        numCols = w.GetColumnCount()
        for i in range(numCols):
            col = w.GetColumn(i)
            col.SetMask(wx.LIST_MASK_TEXT)
            col.SetText(language.translate(self.columns[i]))
            w.SetColumn(i, col)
        
    def addItemWithColumns(self, identifier, *columns):
        w = self.getWxWidget()

        if type(columns[0]) is int:
            # This is an image item.
            index = w.InsertImageItem(w.GetItemCount(), columns[0])
            #info = wx.ListItem()
            #info.m_itemId = index
            #info.m_col = 0
            #info.m_mask = wx.LIST_MASK_FORMAT
            #info.m_format = wx.LIST_FORMAT_CENTRE
            #w.SetItem(info)
        else:
            index = w.InsertStringItem(w.GetItemCount(), uniConv(columns[0]))
            
        for i in range(1, len(columns)):
            w.SetStringItem(index, i, uniConv(columns[i]))

        # Put the identifier into the data.
        w.SetItemData(index, index)

        self.items.append(identifier)

    def setItemImage(self, identifier, image):
        """Sets the image of an image item.
        
        @param identifier  Item identifier.
        @param image  Image index. Needs an image list.
        """
        try:
            index = self.getItemIndex(identifier)
            self.getWxWidget().SetItemImage(index, image)
        except:
            pass

    def sortItems(self, callback):
        self.getWxWidget().SortItems(callback)

    def clear(self):
        """Remove all the items in the list."""

        if self.style in [List.STYLE_MULTIPLE, List.STYLE_SINGLE, List.STYLE_CHECKBOX]:
            self.getWxWidget().Clear()
        else:
            self.getWxWidget().DeleteAllItems()
        self.items = []

    def getItems(self):
        """Returns a list of all the item identifiers in the list."""
        return self.items
        
    def getItemImages(self):
        """Returns a list of all the item iamges in the list. They use the same order
        as the list returned by self.items. Only supported in the Columns style."""
        
        if self.style != List.STYLE_COLUMNS:
            return None
        
        images = [0] * len(self.items)
        
        w = self.getWxWidget()
        for i in range(w.GetItemCount()):
            info = w.GetItem(i, 0)
            images[w.GetItemData(i)] = info.GetImage()
        
        return images
        
    def getSelectedItems(self):
        """Get the identifiers of all the selected items in the list.
        If the list was created with the List.STYLE_CHECKBOX style,
        this returns all the checked items.

        @return An array of identifiers.
        """
        w = self.getWxWidget()
        selected = []

        if self.style == List.STYLE_CHECKBOX:
            # Return a list of checked items.
            for i in range(w.GetCount()):
                if w.IsChecked(i):
                    selected.append(self.items[i])
            return selected

        if self.style == List.STYLE_COLUMNS:
            # Check the state of each item.
            for i in range(w.GetItemCount()):
                state = w.GetItemState(i, wx.LIST_STATE_SELECTED)
                if state & wx.LIST_STATE_SELECTED:
                    selected.append(self.items[w.GetItemData(i)])
            return selected

        # Compose a list of selected items.
        for sel in w.GetSelections():
            selected.append(self.items[sel])
        return selected

    def getSelectedItem(self):
        """Get the identifier of the currently selected item.  Only
        for single-selection lists.

        @return The identifier.
        """
        index = self.getWxWidget().GetSelection()
        if index == wx.NOT_FOUND:
            return ''
        else:
            return self.items[index]

    def getItemIndex(self, identifier):
        """Does the necessary translation for sorted lists."""

        w = self.getWxWidget()
        
        # Unsorted index.
        index = self.items.index(identifier)

        if self.style == List.STYLE_COLUMNS:
            # Column-style lists may be sorted.
            for i in range(w.GetItemCount()):
                if w.GetItemData(i) == index:
                    index = i
                    break
                
        return index

    def selectItem(self, identifier, doSelect=True):
        """Select one item from the list.  If no identifier is
        provided, the selection is cleared."""

        w = self.getWxWidget()

        if identifier == '' or identifier == None:
            # Clear selection.
            if self.style == List.STYLE_COLUMNS:
                # Check the state of each item.
                for i in range(w.GetItemCount()):
                    state = w.GetItemState(i, wx.LIST_STATE_SELECTED)
                    if state & wx.LIST_STATE_SELECTED:
                        w.SetItemState(i, 0, wx.LIST_STATE_SELECTED)
            return
        
        try:
            index = self.getItemIndex(identifier)

            if self.style == List.STYLE_COLUMNS:
                # Check the one item.
                if doSelect:
                    w.SetItemState(index, wx.LIST_STATE_SELECTED,
                                   wx.LIST_STATE_SELECTED)
                else:
                    w.SetItemState(index, 0, wx.LIST_STATE_SELECTED)
            else:
                w.SetSelection(index)
        except:
            pass

    def deselectItem(self, identifier):
        self.selectItem(identifier, False)
        
    def deselectAllItems(self):
        w = self.getWxWidget()
        if self.style == List.STYLE_COLUMNS:
            self.selectItem(None)
        else:
            for i in range(w.GetCount()):
                w.Deselect(i)

    def ensureVisible(self, identifier):
        """Makes sure an item is visible.

        @param identifier  Identifier of the item to make visible.
        """
        w = self.getWxWidget()
        try:
            index = self.getItemIndex(identifier)
            w.EnsureVisible(index)
        except:
            pass
                
    def checkItem(self, identifier, doCheck=True):
        """Check or uncheck an item in a list that was created with
        the Check style.

        @param identifier Identifier of the item to check.

        @param doCheck True, if the item should be checked.  False, if
        unchecked.
        """
        try:
            index = self.getItemIndex(identifier)
            self.getWxWidget().Check(index, doCheck)
        except:
            # Identifier not found?
            pass

    def moveItem(self, identifier, steps):
        """Move an item up or down in the list.

        @param identifier Item to move.

        @param steps Number of indices to move.  Negative numbers move
        the item upwards, positive numbers move it down.
        """
        w = self.getWxWidget()
        
        for index in range(len(self.items)):
            if self.items[index] == identifier:
                # This is the item to move.
                w.Delete(index)
                del self.items[index]

                # Insert it to another location.
                index += steps
                if index < 0:
                    index = 0
                elif index > len(self.items):
                    index = len(self.items)

                w.Insert(language.translate(identifier), index)
                w.SetSelection(index)

                # Also update the items list.
                self.items = self.items[:index] + [identifier] + \
                             self.items[index:]
                break

    def onItemSelected(self, event):
        """Handle the wxWidgets item selection event.

        @param event A wxWidgets event.
        """
        if self.widgetId and self.style == List.STYLE_COLUMNS:
            # Identifier of the selected item is sent with the
            # notification.
            item = self.items[self.getWxWidget().GetItemData(event.GetIndex())]
            
            # Send a notification about the selected item.
            events.send(events.SelectNotify(self.widgetId, item))

        event.Skip()
        self.react()

    def onItemDeselected(self, event):
        """Handle the wxWidgets item deselection event.

        @param event  A wxWidgets event.
        """
        if self.widgetId:
            item = self.items[self.getWxWidget().GetItemData(event.GetIndex())]
            events.send(events.DeselectNotify(self.widgetId, item))

        event.Skip()
        

class DropList (base.Widget):
    """A drop-down list widget."""

    def __init__(self, parent, wxId, id):
        base.Widget.__init__(self, wx.Choice(parent, wxId))

        # Will be used when sending notifications.
        self.widgetId = id

        self.items = []            

        self.addSorted = False

        # We want focus notifications.
        self.setFocusId(id)

        # Listen to profile changes so we can update the selection.
        self.addProfileChangeListener()

        # Handle item selection events.
        wx.EVT_CHOICE(parent, wxId, self.onItemSelected)
        wx.EVT_LEFT_DOWN(self.getWxWidget(), self.onLeftClick)

    def setSorted(self, doSort=True):
        self.addSorted = doSort

    def onLeftClick(self, event):
        if self.widgetId:
            events.send(events.FocusNotify(self.widgetId))
            self.getWxWidget().SetFocus()
        event.Skip()

    def onItemSelected(self, ev):
        """Handle the wxWidgets event that is sent when the current
        selection changes in the list box.

        @param ev A wxWidgets event.
        """
        if self.widgetId:
            newSelection = self.items[ev.GetSelection()]

            if sb.confdb.isSettingDefined(self.widgetId):
                # This is a setting.
                if newSelection == 'default':
                    pr.getActive().removeValue(self.widgetId)
                else:
                    pr.getActive().setValue(self.widgetId, newSelection)

                # Notify everyone of the change.
                events.send(events.EditNotify(self.widgetId, newSelection))
            else:
                # Normal list.
                events.send(events.SelectNotify(self.widgetId, newSelection))

    def onNotify(self, event):
        base.Widget.onNotify(self, event)

        if self.widgetId and sb.confdb.isSettingDefined(self.widgetId):
            if event.hasId('active-profile-changed'):
                # Get the value for the setting as it has been defined
                # in the currently active profile.
                self.getFromProfile(pr.getActive())

    def getFromProfile(self, profile):
        if sb.confdb.isSettingDefined(self.widgetId):
            value = profile.getValue(self.widgetId, False)
            if value:
                self.selectItem(value.getValue())
            else:
                self.selectItem('default')

    def clear(self):
        """Delete all the items."""
        self.items = []
        self.getWxWidget().Clear()

    def __filter(self, itemText):
        """Filter the item text so it can be displayed on screen."""

        # Set the maximum length.
        if len(itemText) > 30:
            return '...' + uniConv(itemText[-30:])

        return uniConv(itemText)

    def retranslate(self):
        """Update the text of the items in the drop list. Preserve
        current selection."""
        
        drop = self.getWxWidget()
        sel = drop.GetSelection()

        # We will replace all the items in the list.
        drop.Clear()        
        for i in range(len(self.items)):
            drop.Append(self.__filter(language.translate(self.items[i])))

        drop.SetSelection(sel)

    def addItem(self, item, toIndex=None):
        """Append a list of items into the drop-down list.

        @param items An array of strings.
        """
        w = self.getWxWidget()
        
        try:
            # Does this already exist?
            index = self.items.index(item)
            # Update label.
            w.SetString(index, language.translate(item))
            return
        except:
            pass
            
        if toIndex is not None:
            # A specific index has been given.
            self.items.insert(toIndex, item)
            w.Insert(language.translate(item), toIndex)
        else:
            if self.addSorted:
                # Need to locate the right index.
                toIndex = 0
                # Sort alphabetically based on identifier.
                for i in range(len(self.items)):
                    if cmp(item, self.items[i]) < 0:
                        break
                    toIndex = i + 1
                self.items.insert(toIndex, item)
                w.Insert(language.translate(item), toIndex)
            else:
                # Just append.
                self.items.append(item)
                w.Append(self.__filter(language.translate(item)))

    def removeItem(self, item):
        try:
            index = self.items.index(item)
            self.getWxWidget().Delete(index)
            del self.items[index]
        except:
            pass

    def hasItem(self, item):
        return item in self.items

    def selectItem(self, item):
        """Change the current selection.  This does not send any
        notifications.

        @param item Identifier of the item to select.
        """
        for i in range(len(self.items)):
            if self.items[i] == item:
                self.getWxWidget().SetSelection(i)
                break

    def getSelectedItem(self):
        """Returns the identifier of the selected item.

        @return An identifier string.
        """
        return self.items[self.getWxWidget().GetSelection()]

    def sortItems(self):
        """Sort all the items in the list."""
        w = self.getWxWidget()
        w.Clear()
        def sorter(a, b):
            if a[0] == 'default':
                return -1
            elif b[0] == 'default':
                return 1
            return cmp(a[1], b[1])
        translated = map(lambda i: (i, language.translate(i)), self.items)
        translated.sort(sorter)
        self.items = []
        for id, item in translated:
            self.items.append(id)
            w.Append(self.__filter(item))

    def getItems(self):
        """Returns the identifiers of the items in the list.

        @return An array of identifiers."""
        return self.items


class HtmlListBox (wx.HtmlListBox):
    """An implementation of the abstract wxWidgets HTML list box."""

    def __init__(self, parent, wxId, itemsListOwner, style=0):
        wx.HtmlListBox.__init__(self, parent, wxId, style=style)

        # A reference to the owner of the array where the items are
        # stored.  Items is a list of tuples: (id, presentation)
        self.itemsListOwner = itemsListOwner

        # Force a resize event at this point so that GTK doesn't
        # select the wrong first visible item.
        #self.SetSize((100, 100))

    def OnGetItem(self, n):
        """Retrieve the representation of an item.  This method must
        be implemented in the derived class and should return the body
        (i.e. without &lt;html&gt; nor &lt;body&gt; tags) of the HTML
        fragment for the given item.

        @param n Item index number.

        @return HTML-formatted contents of the item.
        """
        return self.itemsListOwner.items[n][1]


class FormattedList (base.Widget):
    """A list widget with HTML-formatted items."""

    def __init__(self, parent, wxId, id, style=0):
        """Construct a FormattedList object.

        @param parent Parent wxWidget.
        @param wxId wxWidgets ID of the new widget (integer).
        @param id Identifier of the new widget (string).
        """
        # This item array is also used by the HtmlListBox instance.
        # This is a list of tuples: (id, presentation)
        self.items = []

        base.Widget.__init__(self, HtmlListBox(parent, wxId, self, style=style))

        self.widgetId = id

        # By default, items are added in alphabetical order.
        self.addSorted = True

        # Listen for selection changes.
        wx.EVT_RIGHT_DOWN(self.getWxWidget(), self.onRightDown)      
        wx.EVT_LISTBOX(parent, wxId, self.onItemSelected)
        wx.EVT_LISTBOX_DCLICK(parent, wxId, self.onItemDoubleClick)

    def onRightDown(self, event):
        """Right button down."""

        w = self.getWxWidget()

        #print "right down in list " + str(event.GetPosition())
        item = w.HitTest(event.GetPosition())
        #print item
        if item >= 0:
            w.SetSelection(item)
            if self.widgetId:
                events.send(events.SelectNotify(
                    self.widgetId, self.items[item][0]))

        event.Skip()

    def setSorted(self, doSort):
        self.addSorted = doSort

    def clear(self):
        """Delete all the items."""
        self.items = []
        self.__updateItemCount()

    def addItem(self, itemId, rawItemText, toIndex=None):
        """Add a new item into the formatted list box.  No translation
        is done.  Instead, the provided item text is displayed using
        HTML formatting.

        @param itemId Identifier of the item.  This is sent in
        notification events.

        @param itemText Formatted text for the item.

        @param toIndex Optional index where the item should be placed.
        If None, the item is added to the end of the list.
        """
        itemText = uniConv(rawItemText)
        
        # If this item identifier already exists, update its
        # description without making any further changes.
        for i in range(len(self.items)):
            if self.items[i][0] == itemId:
                self.items[i] = (itemId, itemText)

                # Let the widget know that a change has occured.
                # Although it may be that this does nothing...
                self.__updateItemCount()
                return

        newItem = (itemId, itemText)
        if toIndex == None:
            if self.addSorted:
                toIndex = 0
                # Sort alphabetically based on identifier.
                for i in range(len(self.items)):
                    if cmp(itemId, self.items[i][0]) < 0:
                        break
                    toIndex = i + 1
            else:
                toIndex = len(self.items)

        self.items = self.items[:toIndex] + [newItem] + \
                     self.items[toIndex:]
        self.__updateItemCount()

    def removeItem(self, itemId):
        """Remove an item from the list.

        @param itemId Identifier of the item to remove.
        """
        # Filter out the unwanted item.
        self.items = [item for item in self.items
                      if item[0] != itemId]
        self.__updateItemCount()

    def __updateItemCount(self):
        w = self.getWxWidget()
        w.Hide()
        w.SetItemCount(len(self.items))
        w.SetSelection(self.getSelectedIndex())
        w.Show()

        #w.Freeze()
        #w.SetItemCount(len(self.items))
        #w.SetSelection(self.getSelectedIndex())
        #w.Thaw()
        #w.Refresh()

    def getItemCount(self):
        """Returns the number of items in the list."""
        
        return len(self.items)

    def selectItem(self, itemId):
        """Select an item in the list."""
        for i in range(len(self.items)):
            if self.items[i][0] == itemId:
                self.getWxWidget().SetSelection(i)
                break

    def getSelectedIndex(self):
        count = len(self.items)
        sel = self.getWxWidget().GetSelection()
        if sel >= count:
            sel = count - 1
        return sel

    def getSelectedItem(self):
        """Return the selected item.

        @return Identifier of the selected item.
        """
        return self.items[self.getSelectedIndex()][0]

    def hasItem(self, itemId):
        """Checks if the list contains an item with the specified
        identifier.

        @param itemId Identifier of the item to check.

        @return True, if the item exists.  Otherwise False.
        """
        for item in self.items:
            if item[0] == itemId:
                return True
        return False

    def onItemSelected(self, ev):
        """Handle the wxWidgets event that is sent when the current
        selection changes in the list box.

        @param ev A wxWidgets event.
        """
        ev.Skip()
        events.send(events.SelectNotify(self.widgetId,
                                        self.items[ev.GetSelection()][0]))

    def onItemDoubleClick(self, ev):
        ev.Skip()
        events.send(events.SelectNotify(self.widgetId,
                                        self.items[ev.GetSelection()][0],
                                        True))



# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2005
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file bar.py Bar Widgets

import sys, wx
import host, events, language, paths
import sb.profdb as pr
import base


class Slider (base.Widget):
    def __init__(self, parent, wxId, id):
        """A slider widget for constrained selection of values."""
        base.Widget.__init__(self, wx.Slider(parent, wxId, 50, 0, 100,
                                             style=(wx.SL_AUTOTICKS |
                                                    wx.SL_LABELS |
                                                    wx.SL_HORIZONTAL)))

        self.widgetId = id
        self.step = 1

        # Listen for changes.  Some platform-specific tweaks are needed:
        if host.isUnix():
            wx.EVT_COMMAND_SCROLL_THUMBTRACK(parent, wxId, self.onThumbTrack)
        else:
            wx.EVT_COMMAND_SCROLL_THUMBRELEASE(parent, wxId, self.onThumbTrack)

        # We want focus notifications.
        self.setFocusId(id)

        # Add a command for reseting the slider.
        self.resetCommand = self.widgetId + '-reset-to-default'
        self.setPopupMenu([('reset-slider', self.resetCommand)])
        
        # Listen to our reset command.
        events.addCommandListener(self.onCommand, [self.resetCommand])
        self.addValueChangeListener()
        self.addProfileChangeListener()

        self.oldValue = None

    def setValue(self, value):
        self.getWxWidget().SetValue(value)

    def getValue(self):
        self.snap()
        return self.getWxWidget().GetValue()

    def snap(self):
        w = self.getWxWidget()
        value = w.GetValue()
        w.SetValue(self.step * ((value + self.step/2) / self.step))

    def setRange(self, minimum, maximum, step):
        self.getWxWidget().SetRange(minimum, maximum)

        self.step = step
        if self.step < 1:
            self.step = 1

        self.getWxWidget().SetTickFreq(self.step, 1)
        self.getWxWidget().SetLineSize(self.step)
        self.getWxWidget().SetPageSize(self.step * 2)

    def onThumbTrack(self, ev):
        if self.widgetId:
            newValue = self.getValue()

            if newValue != self.oldValue:
                pr.getActive().setValue(self.widgetId, str(newValue))

                # Send a notification.
                events.send(events.EditNotify(self.widgetId, newValue))
                self.oldValue = newValue

    def __getValueFromProfile(self):
        # Get the value for the setting as it has been defined
        # in the currently active profile.
        value = pr.getActive().getValue(self.widgetId)
        if value:
            self.setValue(int(value.getValue()))

    def onNotify(self, event):
        base.Widget.onNotify(self, event)

        if self.widgetId:
            if (event.hasId('active-profile-changed') or
                event.hasId(self.widgetId + '-value-changed')):
                self.__getValueFromProfile()

    def onCommand(self, event):
        """Handle the slider reset command."""
        
        base.Widget.onCommand(self, event)

        if self.widgetId and event.hasId(self.resetCommand):
            pr.getActive().removeValue(self.widgetId)



#!/usr/bin/env python3

"""Read all test files for a particular file format using a single
importer instance. Read them again in reversed order. This is used
to verify that a loader does proper cleanup and can be called
repeatedly."""

import sys
import os
import subprocess

# hack-load utils.py and settings.py from ../regression
sys.path.append(os.path.join('..','regression'))

import utils
import settings


def process_dir(thisdir):
    """Process /thisdir/ recursively"""
    res = []
    shellparams = {'stdin':subprocess.PIPE,'stdout':sys.stdout,'shell':True}

    command = [utils.assimp_bin_path,"testbatchload"]
    for f in os.listdir(thisdir):
        if os.path.splitext(f)[-1] in settings.exclude_extensions:
            continue
        fullpath = os.path.join(thisdir, f)
        if os.path.isdir(fullpath):
            if f != ".svn":
                res += process_dir(fullpath) 
            continue

        # import twice, importing the same file again introduces extra risk
        # to crash due to garbage data lying around in the importer.
        command.append(fullpath)
        command.append(fullpath)

    if len(command)>2:
        # testbatchload returns always 0 if more than one file in the list worked.
        # however, if it should segfault, the OS will return something not 0.
        command += reversed(command[2:])
        if subprocess.call(command, **shellparams):
            res.append(thisdir)


    return res


def main():
    """Run the test on all registered test repositories"""
    utils.find_assimp_or_die()
    
    res = []
    for tp in settings.model_directories:
        res += process_dir(tp)

    [print(f) for f in res]
    return 0


if __name__ == '__main__':
    res = main()
    input('All done, waiting for keystroke ')

    sys.exit(res)

# vim: ai ts=4 sts=4 et sw=4



#!/usr/bin/env python2.7

import sys, os.path, struct


def split_name(name):
    if '@' in name:
        pos = name.index('@')
        return (name[:pos], name[pos+1:])
    else:
        return (name, name)


def valid_lump_name(name):
    n = name.upper()
    if len(n) > 8:
        # Truncate.
        n = n[:8]
    elif len(n) < 8:
        # Pad with zeroes.
        n += chr(0) * (8 - len(n))
    return n


class Lump:
    def __init__(self, name, data, offset):
        if chr(0) in name:
            name = name[:name.index(chr(0))]
        self.name = name
        self.data = data
        self.offset = offset
        
    def nice_name(self):
        return self.name.strip(chr(0))


class Wad:
    def __init__(self, file_name, mode):
        self.file_name = file_name
        self.mode = mode
        self.file = file(file_name, mode)
        self.type = 'PWAD'
        self.lumps = []
        if 'r' in mode:
            # lumps is a dictionary.
            self.read()
        
    def read(self):
        self.type = self.file.read(4)
        count = struct.unpack('<I', self.file.read(4))[0]
        info_offset = struct.unpack('<I', self.file.read(4))[0]
        print "%s, type=%s, count=%i, info_offset=%i" % (self.file_name, self.type, count, info_offset)
        self.lumps = []
        # Read the info table.
        self.file.seek(info_offset)
        info_table = []
        for i in range(count):
            pos, size = struct.unpack('<II', self.file.read(8))
            name = self.file.read(8)
            info_table.append((pos, size, name))
        # Read the data.
        for info in info_table:
            if info[1] != 0:
                self.file.seek(info[0])
                data = self.file.read(info[1])
            else:
                data = ''
            self.lumps.append( Lump(info[2], data, info[0]) )
        
    def write(self):
        print 'writing', self.file_name
        self.file.truncate()
        self.file.seek(0)
        # Type.
        self.file.write(self.type)
        # Number of lumps and offset.
        self.file.write(struct.pack('<II', len(self.lumps), 0))
        info_table = []
        # Write each lump data.
        for lump in self.lumps:
            size = len(lump.data)
            pos = self.file.tell()
            if size == 0: pos = 0
            info_table.append((pos, size, lump.name))
            if len(lump.data) != 0:
                self.file.write(lump.data)
        
        # Mark the info table offset.
        info_offset = self.file.tell()
        
        # Write the table.
        for info in info_table:
            self.file.write(struct.pack('<II', info[0], info[1]) +
                            valid_lump_name(info[2]))
        
        # Back to write the table offset.
        self.file.seek(8)
        self.file.write(struct.pack('<I', info_offset))
               
    def close(self):
        if 'w' in self.mode or '+' in self.mode:
            self.write()
        self.file.close()
        
    def set_type(self, new_type):
        if len(new_type) != 4:
            print "'%s' is not a valid type. Must have 4 chars." % new_type
        else:
            print 'setting type to %s' % new_type
        self.type = new_type
        
    def list(self):
        for i in range(len(self.lumps)):
            lump = self.lumps[i]
            print "%5i -- %-8s (at %8i, %8i bytes)" % (i, lump.name, lump.offset, 
                                                       len(lump.data))
        print "%i lumps total." % len(self.lumps)
        
    def insert(self, at_lump, names):
        # Where to insert?
        if at_lump is None:
            at = len(self.lumps)
        else:
            at = -1
            for i in range(len(self.lumps)):
                lump = self.lumps[i]
                if lump.nice_name() == at_lump:
                    at = i
                    break
        if at == -1:
            print "could not find insert point, aborting"
            return
        print "inserting at index %i..." % at
        
        for name in names:
            real, arch = split_name(name)
            print "inserting %s as %s" % (real, valid_lump_name(arch))
            self.lumps.insert(at, Lump(valid_lump_name(arch), 
                                       file(real, 'rb').read(), 0))
    
    def extract(self, names):
        split_names = [split_name(n) for n in names]
        names = [arch.upper() for arch, real in split_names]
        for lump in self.lumps:
            nice_name = lump.nice_name()
            if len(names) == 0 or nice_name.upper() in names:
                # Check for a substitution first.
                ext = '.lmp'
                for arch, real in split_names:
                    if arch.upper() == nice_name.upper():
                        if nice_name != real:
                            nice_name = real
                            ext = ''
                        break
                # Write out this lump.
                nice_name = nice_name.lower()
                if os.path.exists(nice_name + ext):
                    counter = 1
                    while True:
                        try_name = nice_name + ext + "_" + str(counter)
                        counter += 1
                        if not os.path.exists(try_name):
                            nice_name = try_name
                            break
                else:
                    nice_name += ext
                print "extracting", lump.name, "(%i bytes) as %s" % (len(lump.data), nice_name)
                f = file(nice_name, 'wb')
                f.write(lump.data)
                f.close()


def usage():
    print 'Usage: %s [command] [wad] [files]' % sys.argv[0]
    print 'command = l(ist) | c(reate PWAD) | C(reate IWAD) | x(tract) | a(ppend) | i(nsert) | t(ype)'
    print 'Examples:'
    print '%s l some.wad' % sys.argv[0]
    print '%s c new.wad LUMP1 LUMP2 LUMP3' % sys.argv[0]
    print '%s i LUMP2 some.wad LUMP5 LUMP6 LUMP7 (at LUMP2)' % sys.argv[0]
    print '%s a existing.wad LUMP4 LUMP5' % sys.argv[0]
    print '%s x existing.wad (extracts all lumps)' % sys.argv[0]
    print '%s x existing.wad LUMP4 (extracts selected lumps)' % sys.argv[0]
    print '%s t existing.wad IWAD' % sys.argv[0]
    print 'You can specify files as realfilename@LUMPNAME.'
   
# Call main function.
print 'WAD Compiler by skyjake@users.sourceforge.net'
print '$Date$'

# Check the args.
if len(sys.argv) < 3:
    usage()
    sys.exit(0)

# Check the command.
command = sys.argv[1][0]
if command not in 'tlcCxai':
    usage()
    sys.exit(1)
     
if command == 'i':
    wad_name = sys.argv[3]
    at_lump = sys.argv[2]
    names = sys.argv[4:]
else:
    wad_name = sys.argv[2]
    names = sys.argv[3:]

# Create a new WAD file.
if command == 'c' or command == 'C':
    wad = Wad(wad_name, 'wb')
    if command == 'C': wad.set_type('IWAD')
    wad.insert(None, names)

else:
    # Open an existing file.
    mode = 'rb'
    if command == 'i' or command == 't': mode = 'rb+'
    try:
        wad = Wad(wad_name, mode)
    except:
        print 'Failure to open %s.' % wad_name
        import traceback
        traceback.print_exc()
        sys.exit(2)
    
    if command == 'l':
        wad.list()
    elif command == 't':
        wad.set_type(names[0])
    elif command == 'i':
        wad.insert(at_lump, names)
    elif command == 'a':
        wad.insert(None, names)
    elif command == 'x':
        wad.extract(names)
    
wad.close()



"""Generates a console help file out of Amethyst sources."""

import os
import sys
import string
import subprocess


def amethyst(input):
    """Runs amethyst with the given input and returns the output."""
    p = subprocess.Popen(['amethyst', '-dHELP'], 
        stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (output, errs) = p.communicate(input)
    #print >>sys.stderr, errs
    return output 
       

def makeHelp(outName, components):
    """Generates the output file 'outName' out of the Amethyst
    sources in the subdirectories listed in 'components'.
    - outName: File path of the output file.
    - components: List of subdirectories to include in the output."""
    
    modes = ['command', 'variable']

    outFile = file(outName, 'w')
    print >> outFile, "# Doomsday Help"
    print >> outFile, "# Generated by conhelp.py out of", string.join(components, ', ')

    for mode in modes:
        for com in components:
            print >> outFile, "#\n# CONSOLE", mode.upper() + 'S: %s\n#' % com
            path = os.path.join(com, mode)
            for fn in os.listdir(path):
                if not fn.endswith('.ame'): continue
                name = fn[:-4]
                print '- %s %s: %s' % (com, mode, name)
                print >> outFile, "\n[%s]" % name
            
                templ = '@require{amestd} @begin\n' + \
                        file(os.path.join(path, fn)).read()
            
                print >> outFile, amethyst(templ).strip()

                
if __name__ == '__main__':
    if len(sys.argv) < 3:
        print "Usage: (outname) (inputdirs) ..."
        sys.exit(0)
    makeHelp(sys.argv[1], sys.argv[2:])



# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2005
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file language.py Localisation

import os, sys, paths, cfparser, events

# This dictionary holds all the known languages.  It maps language
# identifiers to dictionaries of strings.
library = {}

# This dictionary holds all the strings of the current language.  The
# strings may originate from multiple .lang files.
localText = None


def _newString(language, id, text):
    """Insert a new text string into the dictionary.

    @param language The identifier of the language.

    @param id String identifier.

    @param text The translated text.
    """
    #if sys.platform == 'win32':
    #    ## @todo Check this later... Apparently not supported?
    #    text = text.replace('\n', ' ')

    if not library.has_key(language):
        # Start a new language.
        library[language] = {}
        
    library[language][id] = text


def processLanguageBlock(block, prefix=None):
    """Process a block of language strings.  The block must contain
    key elements.

    @param block A cfparser.Block object that contains the language
    strings as cfparser.KeyElement objects.

    @param prefix Prefix to add to the beginning of the string ids.
    """
    for e in block.getContents():
        if e.isKey():
            # The 'category-' strings are NOT affected by the local
            # prefix because they are part of the global information.
            # What a kludge...
            if prefix and e.getName()[:9] != 'category-':
                ident = prefix + '-' + e.getName()
            else:
                ident = e.getName()

            _newString(block.getName(), ident, e.getValue())

            # Define language identifiers always in english, too.
            if ident[:9] == 'language-':
                _newString('english', ident, e.getValue())


def load(fileName):
    """Loads a language file and stores the strings into the database.
    @param fileName Name of the language file.
    """
    p = cfparser.FileParser(fileName)

    # Construct a block element to hold all the strings.
    languageName = paths.getBase(fileName)
    impliedBlock = cfparser.BlockElement('language', languageName)

    try:
        while True:
            # Get the next string from the language file.
            e = p.get()
            
            # Only key elements are recognized.
            if e.isKey():
                impliedBlock.add(e)
    
    except cfparser.OutOfElements:
        # All OK.
        pass

    except:
        import logger
        logger.add(logger.MEDIUM, 'warning-read-language', languageName, 
                   fileName)

    processLanguageBlock(impliedBlock)


def loadAll():
    """Load the <tt>.lang</tt> files from the lang directories.
    """
    for fileName in paths.listFiles(paths.LANG, False):
        # We'll load all the files with the .lang extension.
        if paths.hasExtension('lang', fileName):
            # Load this language file.
            load(fileName)


def loadPath(path):
    """Land all the <tt>.lang</tt> files from the specified directory.

    @param path Directory to read from.
    """
    if not os.path.exists(path):
        return

    for name in os.listdir(path):
        if paths.hasExtension('lang', name):
            load(os.path.join(path, name))


def translate(id, notDefined=None):
    """Look up a translated string from the language database.

    @param id String id.  If this is None or an empty string, an empty
    string is returned.

    @param notDefined The string that is returned if the identifier is
    not defined in the current language or in the default language.
    If None, undefined identifiers are returned untranslated, in
    parentheses.
    """
    # Unspecified IDs are translated to an empty string.
    if not id:
        return ''
    
    try:
        return localText[id]

    except:
        # If the selected language isn't the default (english), see if
        # there is a translation in it.
        if localText != library['english']:
            english = library['english']
            if english.has_key(id):
                return english[id]

        # Return a placeholder.
        if notDefined != None:
            return notDefined
        else:
            return '(' + id + ')'


def define(language, id, text):
    """Define a translation for an identifier.

    @param id String id.

    @param text The translated text.
    """
    _newString(language, id, text)


def isDefined(id):
    """Checks if the string identifier will return a meaningful
    translation.

    @param id String id.
    """
    if localText.has_key(id) or library['english'].has_key(id):
        return True
    
    return False
    

def select(language):
    """Select the language to use for translations.

    @param language The identifier of the language.
    """
    global localText
    localText = library[language]


def change(language):
    """Change the language at runtime.

    @param language Identifier of the new language.
    """
    try:
        if language:
            actual = language
        else:
            actual = 'language-english'

        # Select the new language.
        select(actual[9:])

        # Send a notification so others may update their content.
        events.send(events.LanguageNotify(actual))
        
    except KeyError:
        # Language doesn't exist?
        pass
    

def expand(text, *args):
    """Place the arguments into the text message.

    @param text The text where replacements are done.  A placeholder
    is marked by a percent character and a number (e.g. <tt>%2</tt>).

    @return The expanded version of the string. 
    """
    for i in range(len(args)):
        text = text.replace('%' + str(i + 1), args[i])
    return text.replace('%%', '%')


def getLanguages():
    langs = []
    for fileName in paths.listFiles(paths.LANG):
        # We'll load all the files with the .lang extension.
        if paths.hasExtension('lang', fileName):
            langs.append(paths.getBase(fileName))

    return langs                    


#
# Module Initialization
#

# Strings from the default English language are used for everything by
# default.
loadAll()
select('english')



import os, sys, platform
import shutil
import subprocess
import string
import glob
import gzip
import codecs
import time
import build_number
import config

def remkdir(n):
    if os.path.exists(n):
        print n, 'exists, clearing it...'
        shutil.rmtree(n, True)
    os.mkdir(n)
    
    
def deltree(n):
    if os.path.exists(n):
        shutil.rmtree(n, True)


def omit_path(path, omitted):
    if path.startswith(omitted):
        path = path[len(omitted):]
        if path[0] == '/' or path[0] == '\\': path = path[1:]
    return path


class FileState:
    def __init__(self, isDir, mtime):
        if isDir:
            self.type = 'dir'
        else:
            self.type = 'file'
        self.mtime = int(mtime)
        
    def __repr__(self):
        return "(%s, %i)" % (self.type, self.mtime)


class DirState:
    def __init__(self, path=None, subdirs=True):
        self.files = {} # path -> FileState
        self.subdirs = subdirs
        if path:
            self.update(path, path)
    
    def update(self, path, omitted=None):
        for name in os.listdir(path):
            if name[0] == '.': continue
            fullPath = os.path.join(path, name)
            self.files[omit_path(fullPath, omitted)] = \
                FileState(os.path.isdir(fullPath), os.stat(fullPath).st_mtime)
            if os.path.isdir(fullPath) and self.subdirs:
                self.update(fullPath, omitted)
    
    def list_new_files(self, oldState):
        new = []
        for path in self.files:
            if self.files[path].type == 'dir': continue
            if path not in oldState.files or self.files[path].mtime > oldState.files[path].mtime:
                new.append(path)
        return new
        
    def list_removed(self, oldState):
        """Returns a tuple: (list of removed files, list of removed dirs)"""
        rmFiles = []
        rmDirs = []
        for oldPath in oldState.files:
            if oldPath not in self.files:
                if oldState.files[oldPath].type == 'dir':
                    rmDirs.append(oldPath)
                else:
                    rmFiles.append(oldPath)
        return (rmFiles, rmDirs)


def sys_id():
    bits = platform.architecture()[0]
    
    plat = sys.platform
    
    # Special case: the Snow Leopard builder targets 64-bit.
    if plat == 'darwin':
        macVer = mac_os_version()
        if macVer in ['10.8', '10.9', '10.10', '10.11']:
            plat = 'macx8'
            bits = '64bit'
        elif macVer == '10.6':
            bits = '64bit'
        elif macVer == '10.5':
            plat = 'macx5'
        
    return "%s-%s" % (plat, bits)


def remote_copy(src, dst):
    dst = dst.replace('\\', '/')
    os.system('scp %s %s' % (src, dst))  


def collated(s):
    s = s.strip()
    while s[-1] == '.':
        s = s[:-1]
    return s


def todays_build_tag():
    now = time.localtime()
    return 'build' + build_number.todays_build()


def deb_arch():
    os.system('dpkg --print-architecture > __debarch.tmp')
    arch = file('__debarch.tmp', 'rt').read().strip()
    os.remove('__debarch.tmp')
    return arch


def aptrepo_by_time():
    files = []
    for fn in os.listdir(os.path.join(config.APT_REPO_DIR, 
                                      'dists/unstable/main/binary-' + deb_arch())):
        if fn[-4:] == '.deb':
            files.append(fn)
    return files
    

def aptrepo_find_latest_tag():    
    debs = aptrepo_by_time()
    if not debs: return config.BRANCH
    arch = deb_arch()
    biggest = 0
    for deb in debs:
        number = int(deb[deb.find('-build')+6 : deb.find('_'+arch)])
        biggest = max(biggest, number)
    return 'build' + str(biggest)
    

def count_log_word(fn, word):
    count = 0
    try:
        for txt in [unicode(rl, 'latin1').lower() for rl in string.split(gzip.open(fn).read(), '\n')]:
            pos = txt.find(unicode(word))
            if pos < 0: continue 
            # Ignore some unnecessary gcc messages.
            if 'should be explicitly initialized in the copy constructor' in txt: continue
            if 'deprecated' in txt: continue
            try:
                if txt[pos-1] not in '/\\_'+string.ascii_letters and \
                    txt[pos+len(word)] not in string.ascii_letters+'.' and \
                    txt[pos-11:pos] != 'shlibdeps: ' and txt[pos-12:pos] != 'genchanges: ' and \
                    txt[pos-12:pos] != 'cc1objplus: ':
                    count += 1            
            except IndexError:
                count += 1
    except:
        pass
    return count


def count_log_issues(fn):
    """Returns tuple of (#warnings, #errors) in the fn."""
    return (count_log_word(fn, 'error'), count_log_word(fn, 'warning'))    


def count_word(word, inText):
    pos = 0
    count = 0
    while True:
        pos = inText.find(word, pos)
        if pos < 0: break
        count += 1
        pos += len(word)
    return count


def mac_os_version():
    """Determines the Mac OS version."""
    ver = platform.mac_ver()[0]
    if not ver: return None
    if ver.count('.') == 1: # "10.9"
        return ver
    return ver[:ver.rindex('.')] # "10.9.3"


def version_split(versionText):
    return map(int, versionText.split('.'))


def version_cmp(a, b):
    """Compares two versions, returning -1 if a < b, 0 if a == b, and 1 if a > b.
    
    Arguments:
        a:  String in the form "1.2.3"
        b:  String in the form "3.4.5"
    """
    va = version_split(a)
    vb = version_split(b)
    if va < vb: return -1
    if va > vb: return 1
    return 0
    
    
def system_command(cmd):
    result = subprocess.call(cmd, shell=True)
    if result != 0:
        raise Exception("System command \"%s\" returned error code %i" % (cmd, result))


def python2_executable():
    if sys.platform[:3] == 'win':
        return 'python'
    elif mac_os_version() == '10.5':
        return '/usr/bin/env python2.5'  # required by Snowberry
    else:
        return '/usr/bin/env python2.7'


def run_python2(script):
    system_command(python2_executable() + " " + script)



#!/usr/bin/env python3
# -*- Coding: UTF-8 -*-

# ---------------------------------------------------------------------------
# Open Asset Import Library (ASSIMP)
# ---------------------------------------------------------------------------
#
# Copyright (c) 2006-2010, ASSIMP Development Team
#
# All rights reserved.
#
# Redistribution and use of this software in source and binary forms, 
# with or without modification, are permitted provided that the following 
# conditions are met:
# 
# * Redistributions of source code must retain the above
#   copyright notice, this list of conditions and the
#   following disclaimer.
# 
# * Redistributions in binary form must reproduce the above
#   copyright notice, this list of conditions and the
#   following disclaimer in the documentation and/or other
#   materials provided with the distribution.
# 
# * Neither the name of the ASSIMP team, nor the names of its
#   contributors may be used to endorse or promote products
#   derived from this software without specific prior
#   written permission of the ASSIMP Development Team.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# ---------------------------------------------------------------------------

"""Parse an EXPRESS file and extract basic information on all
entities and data types contained"""

import sys, os, re

re_match_entity = re.compile(r"""
ENTITY\s+(\w+)\s*                                    # 'ENTITY foo'
.*?                                                  #  skip SUPERTYPE-of
(?:SUBTYPE\s+OF\s+\((\w+)\))?;                       # 'SUBTYPE OF (bar);' or simply ';'
(.*?)                                                # 'a : atype;' (0 or more lines like this)
(?:(?:INVERSE|UNIQUE|WHERE)\s*$.*?)?                 #  skip the INVERSE, UNIQUE, WHERE clauses and everything behind 
END_ENTITY;                                          
""",re.VERBOSE|re.DOTALL|re.MULTILINE) 

re_match_type = re.compile(r"""
TYPE\s+(\w+?)\s*=\s*((?:LIST|SET)\s*\[\d+:[\d?]+\]\s*OF)?(?:\s*UNIQUE)?\s*(\w+)   # TYPE foo = LIST[1:2] of blub
(?:(?<=ENUMERATION)\s*OF\s*\((.*?)\))?
.*?                                                                 #  skip the WHERE clause
END_TYPE;
""",re.VERBOSE|re.DOTALL)

re_match_field = re.compile(r"""
\s+(\w+?)\s*:\s*(OPTIONAL)?\s*((?:LIST|SET)\s*\[\d+:[\d?]+\]\s*OF)?(?:\s*UNIQUE)?\s*(\w+?);
""",re.VERBOSE|re.DOTALL)


class Schema:
    def __init__(self):
        self.entities = {}
        self.types = {}

class Entity:
    def __init__(self,name,parent,members):
        self.name = name
        self.parent = parent
        self.members = members

class Field:
    def __init__(self,name,type,optional,collection):
        self.name = name
        self.type = type
        self.optional = optional
        self.collection = collection
        self.fullspec = (self.collection+' ' if self.collection else '') + self.type 

class Type:
    def __init__(self,name,aggregate,equals,enums):
        self.name = name
        self.aggregate = aggregate
        self.equals = equals
        self.enums = enums
        

def read(filename,silent=False):
    schema = Schema()
    with open(filename,'rt') as inp: 
        contents = inp.read()
        types = re.findall(re_match_type,contents)
        for name,aggregate,equals,enums in types:
            schema.types[name] = Type(name,aggregate,equals,enums)
            
        entities = re.findall(re_match_entity,contents)
        for name,parent,fields_raw in entities:
            print('process entity {0}, parent is {1}'.format(name,parent)) if not silent else None
            fields = re.findall(re_match_field,fields_raw)
            members = [Field(name,type,opt,coll) for name, opt, coll, type in fields]
            print('  got {0} fields'.format(len(members))) if not silent else None
            
            schema.entities[name] = Entity(name,parent,members)
    return schema

if __name__ == "__main__":
    sys.exit(read(sys.argv[1] if len(sys.argv)>1 else 'schema.exp'))




    



#!/usr/bin/env python3
# -*- Coding: UTF-8 -*-

# ---------------------------------------------------------------------------
# Open Asset Import Library (ASSIMP)
# ---------------------------------------------------------------------------
#
# Copyright (c) 2006-2010, ASSIMP Development Team
#
# All rights reserved.
#
# Redistribution and use of this software in source and binary forms, 
# with or without modification, are permitted provided that the following 
# conditions are met:
# 
# * Redistributions of source code must retain the above
#   copyright notice, this list of conditions and the
#   following disclaimer.
# 
# * Redistributions in binary form must reproduce the above
#   copyright notice, this list of conditions and the
#   following disclaimer in the documentation and/or other
#   materials provided with the distribution.
# 
# * Neither the name of the ASSIMP team, nor the names of its
#   contributors may be used to endorse or promote products
#   derived from this software without specific prior
#   written permission of the ASSIMP Development Team.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# ---------------------------------------------------------------------------

"""Shared settings for the regression suite (bold builder and
test scripts rely on this)

"""

import os

# -------------------------------------------------------------------------------
# List of file extensions to be excluded from the regression suite
# File extensions are case insensitive
# -------------------------------------------------------------------------------
exclude_extensions = [
    ".lws", ".assbin", ".assxml", ".txt", ".md",
    ".jpeg", ".jpg", ".png", ".gif", ".tga", ".bmp",
    ".skeleton", ".skeleton.xml"
]

# -------------------------------------------------------------------------------
# Post processing configurations to be included in the test. The
# strings are parameters for assimp_cmd, see assimp_cmd's doxydoc
# for more details. 

# The defaults are (validate-data-structure is always enabled, for
# self-explanatory reasons :-):
#
# '-cfull'    :apply all post processing except 'og' and 'ptv' (optimize-scenegraph)
# '-og -om'   :run optimize-scenegraph in combination with optimize-meshes.
# '-vds -jiv' :join-identical-vertices alone. This is a hotspot where
#              floating-point inaccuracies can cause severe damage.
# '-ptv':      transform all meshes to world-space

# As you can see, not all possible combinations of pp steps are covered - 
# but at least each step is executed at least once on each model. 
# -------------------------------------------------------------------------------
pp_configs_to_test = [
    "-cfull",
    "-og -om -vds",
    "-vds -jiv",
    "-ptv -gsn -cts -db",

    # this is especially important: if no failures are present with this 
    # preset, the regression is most likely caused by the post
    # processing pipeline.
    ""
]
# -------------------------------------------------------------------------------
# Name of the regression database file to be used
# gen_db.py writes to this directory, run.py checks against this directory.
# If a zip file with the same name exists, its contents are favoured to a 
# normal directory, so in order to test against unzipped files the ZIP needs
# to be deleted.
# -------------------------------------------------------------------------------
database_name = "db"

# -------------------------------------------------------------------------------
# List of directories to be processed. Paths are processed recursively.
# -------------------------------------------------------------------------------
model_directories = [
    os.path.join("..","models"),
    os.path.join("..","models-nonbsd")
]

# -------------------------------------------------------------------------------
# Remove the original database files after the ZIP has been built?
# -------------------------------------------------------------------------------
remove_old = True

# -------------------------------------------------------------------------------
# Bytes to skip at the beginning of a dump. This skips the file header, which
# is currently the same 500 bytes header for both assbin, assxml and minidumps.
# -------------------------------------------------------------------------------
dump_header_skip = 500

# -------------------------------------------------------------------------------
# Directory to write all results and logs to. The dumps pertaining to failed
# tests are written to a subfolder of this directory ('tmp').
# -------------------------------------------------------------------------------
results = os.path.join("..","results")

# Create results directory if it does not exist
if not os.path.exists(results):
    os.makedirs(results)

# vim: ai ts=4 sts=4 et sw=4



#!/usr/bin/env python3
# -*- Coding: UTF-8 -*-

# ---------------------------------------------------------------------------
# Open Asset Import Library (ASSIMP)
# ---------------------------------------------------------------------------
#
# Copyright (c) 2006-2010, ASSIMP Development Team
#
# All rights reserved.
#
# Redistribution and use of this software in source and binary forms, 
# with or without modification, are permitted provided that the following 
# conditions are met:
# 
# * Redistributions of source code must retain the above
#   copyright notice, this list of conditions and the
#   following disclaimer.
# 
# * Redistributions in binary form must reproduce the above
#   copyright notice, this list of conditions and the
#   following disclaimer in the documentation and/or other
#   materials provided with the distribution.
# 
# * Neither the name of the ASSIMP team, nor the names of its
#   contributors may be used to endorse or promote products
#   derived from this software without specific prior
#   written permission of the ASSIMP Development Team.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# ---------------------------------------------------------------------------

"""Generate the C++ glue code needed to map EXPRESS to C++"""

import sys, os, re

if sys.version_info < (3, 0):
    print("must use python 3.0 or greater")
    sys.exit(-2)

input_template_h = 'IFCReaderGen.h.template'
input_template_cpp = 'IFCReaderGen.cpp.template'

output_file_h = os.path.join('..','..','code','IFCReaderGen.h')
output_file_cpp = os.path.join('..','..','code','IFCReaderGen.cpp')

template_entity_predef = '\tstruct {entity};\n'
template_entity_predef_ni = '\ttypedef NotImplemented {entity}; // (not currently used by Assimp)\n'
template_entity = r"""

    // C++ wrapper for {entity}
    struct {entity} : {parent} ObjectHelper<{entity},{argcnt}> {{ {entity}() : Object("{entity}") {{}}
{fields}
    }};"""

template_entity_ni = ''

template_type = r"""
    // C++ wrapper type for {type}
    typedef {real_type} {type};"""

template_stub_decl = '\tDECL_CONV_STUB({type});\n'
template_schema = '\t\tSchemaEntry("{normalized_name}",&STEP::ObjectHelper<{type},{argcnt}>::Construct )\n'
template_schema_type = '\t\tSchemaEntry("{normalized_name}",NULL )\n'
template_converter = r"""
// -----------------------------------------------------------------------------------------------------------
template <> size_t GenericFill<{type}>(const DB& db, const LIST& params, {type}* in)
{{
{contents}
}}"""

template_converter_prologue_a = '\tsize_t base = GenericFill(db,params,static_cast<{parent}*>(in));\n'
template_converter_prologue_b = '\tsize_t base = 0;\n'
template_converter_check_argcnt = '\tif (params.GetSize() < {max_arg}) {{ throw STEP::TypeError("expected {max_arg} arguments to {name}"); }}'
template_converter_code_per_field = r"""    do {{ // convert the '{fieldname}' argument
        boost::shared_ptr<const DataType> arg = params[base++];{handle_unset}{convert}
    }} while(0);
"""
template_allow_optional = r"""
        if (dynamic_cast<const UNSET*>(&*arg)) break;"""
template_allow_derived = r"""
        if (dynamic_cast<const ISDERIVED*>(&*arg)) {{ in->ObjectHelper<Assimp::IFC::{type},{argcnt}>::aux_is_derived[{argnum}]=true; break; }}"""        
template_convert_single = r"""
        try {{ GenericConvert( in->{name}, arg, db ); break; }} 
        catch (const TypeError& t) {{ throw TypeError(t.what() + std::string(" - expected argument {argnum} to {classname} to be a `{full_type}`")); }}"""

template_converter_omitted = '// this data structure is not used yet, so there is no code generated to fill its members\n'
template_converter_epilogue = '\treturn base;'

import ExpressReader


def get_list_bounds(collection_spec):
    start,end = [(int(n) if n!='?' else 0) for n in re.findall(r'(\d+|\?)',collection_spec)]
    return start,end

def get_cpp_type(field,schema):
    isobjref = field.type in schema.entities
    base = field.type
    if isobjref:
        base = 'Lazy< '+(base if base in schema.whitelist else 'NotImplemented')+' >'
    if field.collection:
        start,end = get_list_bounds(field.collection)
        base = 'ListOf< {0}, {1}, {2} >'.format(base,start,end)
    if not isobjref:
        base += '::Out'
    if field.optional:
        base = 'Maybe< '+base+' >'
        
    return base

def generate_fields(entity,schema):
    fields = []
    for e in entity.members:
        fields.append('\t\t{type} {name};'.format(type=get_cpp_type(e,schema),name=e.name))
    return '\n'.join(fields)

def handle_unset_args(field,entity,schema,argnum):
    n = ''
    # if someone derives from this class, check for derived fields.
    if any(entity.name==e.parent for e in schema.entities.values()):
        n += template_allow_derived.format(type=entity.name,argcnt=len(entity.members),argnum=argnum)
    
    if not field.optional:
        return n+''
    return n+template_allow_optional.format()

def get_single_conversion(field,schema,argnum=0,classname='?'):
    typen = field.type
    name = field.name
    if field.collection:
        typen = 'LIST'
    return template_convert_single.format(type=typen,name=name,argnum=argnum,classname=classname,full_type=field.fullspec)

def count_args_up(entity,schema):
    return len(entity.members) + (count_args_up(schema.entities[entity.parent],schema) if entity.parent else 0)

def resolve_base_type(base,schema):
    if base in ('INTEGER','REAL','STRING','ENUMERATION','BOOLEAN','NUMBER', 'SELECT','LOGICAL'):
        return base
    if base in schema.types:
        return resolve_base_type(schema.types[base].equals,schema)
    print(base)
    return None
    
def gen_type_struct(typen,schema):
    base = resolve_base_type(typen.equals,schema)
    if not base:
        return ''

    if typen.aggregate:
        start,end = get_list_bounds(typen.aggregate)
        base = 'ListOf< {0}, {1}, {2} >'.format(base,start,end)
    
    return template_type.format(type=typen.name,real_type=base)

def gen_converter(entity,schema):
    max_arg = count_args_up(entity,schema)
    arg_idx = arg_idx_ofs = max_arg - len(entity.members)
    
    code = template_converter_prologue_a.format(parent=entity.parent) if entity.parent else template_converter_prologue_b
    if entity.name in schema.blacklist_partial:
        return code+template_converter_omitted+template_converter_epilogue;
        
    if max_arg > 0:
        code +=template_converter_check_argcnt.format(max_arg=max_arg,name=entity.name)

    for field in entity.members:
        code += template_converter_code_per_field.format(fieldname=field.name,
            handle_unset=handle_unset_args(field,entity,schema,arg_idx-arg_idx_ofs),
            convert=get_single_conversion(field,schema,arg_idx,entity.name))

        arg_idx += 1
    return code+template_converter_epilogue

def get_base_classes(e,schema):
    def addit(e,out):
        if e.parent:
            out.append(e.parent)
            addit(schema.entities[e.parent],out)
    res = []
    addit(e,res)
    return list(reversed(res))

def get_derived(e,schema):
    def get_deriv(e,out): # bit slow, but doesn't matter here
        s = [ee for ee in schema.entities.values() if ee.parent == e.name]
        for sel in s:
            out.append(sel.name)
            get_deriv(sel,out)
    res = []
    get_deriv(e,res)
    return res

def get_hierarchy(e,schema):
    return get_derived(e.schema)+[e.name]+get_base_classes(e,schema)

def sort_entity_list(schema):
    deps = []
    entities = schema.entities
    for e in entities.values():
        deps += get_base_classes(e,schema)+[e.name]

    checked = [] 
    for e in deps: 
        if e not in checked: 
            checked.append(e)
    return [entities[e] for e in checked]
    
def work(filename):
    schema = ExpressReader.read(filename,silent=True)
    entities, stub_decls, schema_table, converters, typedefs, predefs = '','',[],'','',''

    
    whitelist = []
    with open('entitylist.txt', 'rt') as inp:
        whitelist = [n.strip() for n in inp.read().split('\n') if n[:1]!='#' and n.strip()]

    schema.whitelist = set()
    schema.blacklist_partial = set()
    for ename in whitelist:
        try:
            e = schema.entities[ename]
        except KeyError:
            # type, not entity
            continue
        for base in [e.name]+get_base_classes(e,schema):
            schema.whitelist.add(base)
        for base in get_derived(e,schema):
            schema.blacklist_partial.add(base)

    schema.blacklist_partial -= schema.whitelist
    schema.whitelist |= schema.blacklist_partial

    # uncomment this to disable automatic code reduction based on whitelisting all used entities
    # (blacklisted entities are those who are in the whitelist and may be instanced, but will
    # only be accessed through a pointer to a base-class.
    #schema.whitelist = set(schema.entities.keys())
    #schema.blacklist_partial = set()

    for ntype in schema.types.values():
        typedefs += gen_type_struct(ntype,schema)
        schema_table.append(template_schema_type.format(normalized_name=ntype.name.lower()))

    sorted_entities = sort_entity_list(schema)
    for entity in sorted_entities:
        parent = entity.parent+',' if entity.parent else ''

        if entity.name in schema.whitelist:
            converters += template_converter.format(type=entity.name,contents=gen_converter(entity,schema))
            schema_table.append(template_schema.format(type=entity.name,normalized_name=entity.name.lower(),argcnt=len(entity.members)))
            entities += template_entity.format(entity=entity.name,argcnt=len(entity.members),parent=parent,fields=generate_fields(entity,schema))
            predefs += template_entity_predef.format(entity=entity.name)
            stub_decls += template_stub_decl.format(type=entity.name)
        else:
            entities += template_entity_ni.format(entity=entity.name)
            predefs += template_entity_predef_ni.format(entity=entity.name)
            schema_table.append(template_schema.format(type="NotImplemented",normalized_name=entity.name.lower(),argcnt=0))

    schema_table = ','.join(schema_table)

    with open(input_template_h,'rt') as inp:
        with open(output_file_h,'wt') as outp:
            # can't use format() here since the C++ code templates contain single, unescaped curly brackets
            outp.write(inp.read().replace('{predefs}',predefs).replace('{types}',typedefs).replace('{entities}',entities).replace('{converter-decl}',stub_decls))
    
    with open(input_template_cpp,'rt') as inp:
        with open(output_file_cpp,'wt') as outp:
            outp.write(inp.read().replace('{schema-static-table}',schema_table).replace('{converter-impl}',converters))

if __name__ == "__main__":
    sys.exit(work(sys.argv[1] if len(sys.argv)>1 else 'schema.exp'))



    
    



# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2006
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file conf.py Configuration Setting Classes
## 
## This module contains the classes that manage all the available
## settings and components.  The actual configuration with which
## Doomsday is launched is composed of the data in the settings and
## components.

# Import modules required for initialization.
import os, re, string
import paths
import sb.expressions as ex


# Setting creation order counter.
creationOrder = 1


class Component:
    """Components are dynamically loaded libraries that Doomsday
    loads.  Each component implements certain functionality."""
    
    def __init__(self, id, setting, option=''):
        self.id = id
        self.option = option
        self.setting = setting

    def getId(self):
        """Return the ID of the Component."""
        return self.id

    def getSetting(self):
        """Return the identifier of the setting that selects this
        component.
        """
        return self.setting

    def getOption(self):
        """Returns extra options to be used with the Component."""
        return ex.evaluate(self.option, None)


class Setting:
    """Setting is the abstract base class for all the settings."""    
    
    def __init__(self, type, id, option):
        self.type = type
        self.id = id
        self.option = option

        # Append value after option on the command line.
        self.valueAfterOption = True

        # Settings will appear in the order they have been created.
        global creationOrder
        self.order = creationOrder
        creationOrder += 1

        self.default = None

        # Group determines the tab on which the setting will be
        # displayed in Settings.
        self.group = None

        # Subgroup is an optional extra grouping within the setting
        # tab.
        self.subGroup = None

        # An array of tuples: (setting-id, value).
        self.requiredValues = []

        # Lists of required components and addons (identifiers).
        self.components = []
        self.addons = []            

    def getId(self):
        """Return the ID of the Setting."""
        return self.id

    def getOption(self):
        """Return the command line option of the Setting."""
        return self.option

    def getOrder(self):
        """Return the creation order number of the Setting.

        @return An integer.
        """
        return self.order
    
    def getType(self):
        """Returns the type of the Setting.
        @return Text string that is one of 'toggle', 'range', etc.
        """
        return self.type

    def getDefault(self):
        """Returns the default value for the Setting, if one has been
        defined.

        @param The default value.  May be None.
        """
        return self.default

    def getRequiredComponents(self):
        """Returns a listing of the components required by the setting.

        @return An array of string identifiers.
        """
        return self.components

    def getRequiredAddons(self):
        """Returns a listing of the addons required by the setting.

        @return An array of string identifiers.
        """
        return self.addons

    def getRequiredValues(self):
        """@return An array of tuples: (setting-id, value)
        """
        return self.requiredValues

    def getOption(self):
        return self.option

    def setGroup(self, group):
        self.group = group

    def setSubGroup(self, subGroup):
        self.subGroup = subGroup

    def getGroup(self):
        return self.group

    def getSubGroup(self):
        return self.subGroup

    def setValueAfterOption(self, doAppend):
        self.valueAfterOption = (doAppend == 'yes')

    def isValueAfterOption(self):
        return self.valueAfterOption

    def addRequiredComponent(self, component):
        if component:
            self.components.append(component)

    def addRequiredAddon(self, addon):
        if addon:
            self.addons.append(addon)

    def addRequiredValue(self, settingId, value):
        self.requiredValues.append((settingId, value))

    def isEnabled(self, profile):
        """Checks if the setting can be used based on the current
        values of the profile.  Some settings conflict each other.
        This method decides which settings will actually be used for
        launching the game.
        """
        # Check that all the required components are in use.
        if len(self.components) > 0:
            profComps = profile.getComponents()
            for c in self.components:
                if c not in profComps:
                    return False
            
        for settingId, requiredValue in self.getRequiredValues():
            v = profile.getValue(settingId)

            # If there is no value and something is required, the
            # requirement isn't met.
            if not v and requiredValue != '':
                return False

            # If the setting has the wrong value, the requirement
            # isn't met.
            if v.getValue() != requiredValue:
                return False

        return True
        
    def getCommandLine(self, profile):
        """Form the command line options that represent the value of
        the setting.  The value is taken from the specified profile.
        If the value is unspecified, no command line option is
        generated.  This happens when the Defaults profile doesn't
        have a default for the setting.

        @param profile A profiles.Profile object.  The values of the
        settings will be taken from this profile.

        @return A string that contains all the necessary command line
        options.
        """
        cmdLine = self.composeCommandLine(profile)

        # Evaluate expressions in the command line.
        return ex.evaluate(cmdLine, self.getContextAddon())

    def composeCommandLine(self, profile):
        """Overridden by subclasses."""
        return self.getOption()

    def getContextAddon(self):
        """Determines the addon with which the setting is associated.
        If the setting is not associated with an addon, return None.

        @return  Addon object.
        """
        if len(self.addons) > 0:
            # This is hardcoded to take the first addon.  Settings
            # don't need to be bound to just one addon...!
            import sb.aodb
            return sb.aodb.get(self.addons[0])
        else:
            return None                
    

class ToggleSetting (Setting):
    def __init__(self, id, option, default, inactiveOption=''):
        Setting.__init__(self, 'toggle', id, option)
        if default == 'yes':
            self.default = default
        else:
            self.default = 'no'
        self.inactiveOption = inactiveOption

    def composeCommandLine(self, profile):
        value = profile.getValue(self.getId())
        if value:
            if value.getValue() == 'yes':
                return self.getOption()
            elif self.inactiveOption:
                return self.inactiveOption
        return ""
        

class RangeSetting (Setting):
    def __init__(self, id, option, minimum, maximum, default, suffix=''):
        Setting.__init__(self, 'range', id, option)
        self.min = int(minimum)
        self.max = int(maximum)
        self.default = int(default)
        self.suffix = suffix

    def getMinimum(self):
        return self.min

    def getMaximum(self):
        return self.max
    
    def composeCommandLine(self, profile):
        value = profile.getValue(self.getId())
        if value:
            if self.isValueAfterOption():
                return self.getOption() + ' ' + value.getValue() + self.suffix
            else:
                # No value after option.
                return self.getOption() 
        return ""


class SliderSetting (RangeSetting):
    """A RangeSetting with a step constraint."""

    def __init__(self, id, option, minimum, maximum, step, default, suffix=''):
        RangeSetting.__init__(self, id, option, minimum, maximum,
                              default, suffix)
        self.type = 'slider'
        self.step = int(step)

    def getStep(self):
        return self.step


class ChoiceSetting (Setting):
    def __init__(self, id, option, alternatives, cmdLines, default):
        """Construct a ChoiceSetting.

        @param id Identifier of the setting (string).

        @param option Command line option associated with this setting.

        @param alternatives Array of strings, representing the
        identifiers of each choice.

        @param cmdLines Array of strings, containing the command-line
        equivalent of each choice.

        @param default Identifier of the default choice.
        """
        Setting.__init__(self, 'choice', id, option)
        # The identifiers of the choices.  Add the setting identifier
        # to the beginning of each identifier.
        self.alternatives = map(lambda alt: self.getId() + '-' + alt,
                                alternatives)
        # The command line equivalents of each choice.
        self.cmdLines = cmdLines

        self.sortChoices = False

        # Identifier of the default choice.
        if len(default) > 0:
            self.default = self.getId() + '-' + default
        
    def getAlternatives(self):
        """Returns the list of possible choices.

        @return Array of strings, containing the identifiers of all
        the choices.
        """
        return self.alternatives

    def composeCommandLine(self, profile):
        value = profile.getValue(self.getId())
        if value:
            # Find the corresponding command line parameter.
            try:
                index = self.alternatives.index(value.getValue())
                return self.getOption() + ' ' + self.cmdLines[index]
            except:
                # Unknown choice, how strange.
                pass
        return ""

    def isSorted(self):
        return self.sortChoices
   
    def setSorted(self, shouldSort):
        self.sortChoices = shouldSort
        
    def merge(self, alts, opts):
        """Merges the alternatives and command lines of the other choice
        setting with this one. Merging means that duplicates are not added."""
        
        for alt in alts:
            if alt not in self.alternatives:
                self.alternatives.append(alt)
        for cmd in opts:
            if cmd not in self.cmdLines:
                self.cmdLines.append(cmd)


class FileSetting (Setting):
    """A file name setting.  If the file must exist, it is selected
    using a standard file selection dialog.  If not, the name can be
    entered in a text field."""

    def __init__(self, id, option, mustExist, allowedTypes, default):
        Setting.__init__(self, 'file', id, option)
        self.mustExist = (mustExist == 'yes')
        if len(default) > 0:
            self.default = default
        self.allowedTypes = allowedTypes

    def hasToExist(self):
        return self.mustExist

    def getAllowedTypes(self):
        """@return Allowed types as a an array of tuples (name, extension).
        """
        return self.allowedTypes

    def composeCommandLine(self, profile):
        value = profile.getValue(self.getId())
        if value:
            if self.isValueAfterOption():
                v = self.getOption() + ' $PATH<' + value.getValue() + '>'
            else:
                v = self.getOption()
            return v
        return ""


class FolderSetting (FileSetting):
    """A folder name setting, selected using a standard folder
    selection dialog."""

    def __init__(self, id, option, mustExist, default):
        FileSetting.__init__(self, id, option, mustExist, [], default)
        self.type = 'folder'


class TextSetting (Setting):
    """A text setting with an arbitrary text string that is added
    unmodified to the command line."""

    def __init__(self, id, option, default):
        Setting.__init__(self, 'text', id, option)
        if len(default):
            self.default = default

    def composeCommandLine(self, profile):
        value = profile.getValue(self.getId())
        if value:
            if self.getOption() != '':
                # An option has been defined for the setting.
                if self.isValueAfterOption():
                    v = self.getOption() + ' ' + paths.quote(value.getValue())
                else:
                    v = self.getOption()
            else:
                # No option has been defined, so just add the text as-is.
                v = value.getValue()
            return v
        return ""



# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2011
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file launcher.py Launcher Plugin
##
## Implements the user interface for launching the game.  Handles
## confliction resolution using dialogs.
##
## A cleaner implementation might define its own dialog class for the
## resolver, derived from AreaDialog.

# TODO: Implement a timer service so wx isn't needed here.

import os, string, time, subprocess
import ui, host
import sb.util.dialog
import events
import paths
import language
import sb.widget.button as wg
import sb.profdb as pr
import sb.confdb as st
import sb.aodb as ao
import sb.expressions as ex


class MessageTimer (ui.Timer):
    """The message timer will clear the launch message when the 
    timer expires."""

    def __init__(self):
        ui.Timer.__init__(self)
        
    def expire(self):
        """Called when the timer expires."""
        
        # Clear the launch message.
        setLaunchMessage('')


# An instance of the message timer.
launchTextTimer = MessageTimer()


class ResolveStatus:
    def __init__(self):
        self.dialog = None
        self.area = None
        self.problem = None
        self.profile = None
        
    def dontUse(self, addon):
        self.profile.dontUseAddon(addon.getId())
        # Also remove from the set of addons.
        if addon in self.addons:
            self.addons.remove(addon)

    def createConflictButtons(self, addonName):
        self.area.setWeight(0)
        self.choices = []
        self.choices.append(
            resolving.area.createRadioButton('', True, True))
        self.choices.append(
            resolving.area.createRadioButton('', False))
            
        self.choices[0].setText(language.expand(
            language.translate('conflict-choice1'), addonName))
            
        self.choices[1].setText(language.expand(
            language.translate('conflict-choice2'), addonName))


def init():
    """Create the Play button."""
    
    area = ui.getArea(ui.COMMAND)
    area.addSpacer()

    global logFileName
    logFileName = os.path.join(paths.getUserPath(paths.RUNTIME),
                               "Conflicts.log")

    global launchText
    launchText = area.createText('') #launch-message')

    global playButton
    playButton = area.createButton('play', wg.Button.STYLE_DEFAULT)

    global resolving
    resolving = ResolveStatus()

    playButton.setPopupMenu(['view-command-line'])

    # Register for listening to notifications.
    events.addNotifyListener(handleNotify, ['active-profile-changed'])

    # Register for listening to commands.
    events.addCommandListener(handleCommand, ['play', 'view-command-line',
                                              'continue'])
    
    # Commands for the popup menu.
    ui.addMenuCommand(ui.MENU_APP, 'quit', group=ui.MENU_GROUP_APP)
    ui.addMenuCommand(ui.MENU_PROFILE, 'play', group=ui.MENU_GROUP_LAUNCH)
    ui.addMenuCommand(ui.MENU_TOOLS, 'view-command-line', group=ui.MENU_GROUP_LAUNCH)


def handleNotify(event):
    if event.hasId('active-profile-changed'):
        # Disable or enable controls based on which profile is selected.
        if pr.getActive() is pr.getDefaults():
            playButton.disable()
            ui.disableMenuCommand('play')
            ui.disableMenuCommand('view-command-line')
        else:
            playButton.enable()
            ui.enableMenuCommand('play')
            ui.enableMenuCommand('view-command-line')


def handleCommand(event):
    if pr.getActive() is pr.getDefaults():
        return

    if event.hasId('play'):
        # Launch the game with the active profile.
        startGame(pr.getActive())

    elif event.hasId('continue') and resolving:
        continueResolving()

    elif event.hasId('view-command-line'):
        # Generate a command line and display it in a dialog.
        options = generateOptions(pr.getActive())
        if options == None:
            return

        # Escape any angle brackets.
        options = options.replace('<', '&lt;')
        options = options.replace('>', '&gt;')
        options = options.replace(' -', '<br>-')

        # Highlight all the options with a bold font.
        pos = 0
        while pos < len(options):
            if options[pos:pos + 5] == '<br>-' or \
                   pos == 0 and options[pos] == '-':
                options = options[:pos] + '<b>' + options[pos:]
                pos += 5
                while pos < len(options) and \
                      options[pos] not in string.whitespace:
                    pos += 1
                options = options[:pos] + '</b>' + options[pos:]
            pos += 1

        dialog, area = sb.util.dialog.createButtonDialog(
            'view-command-line-dialog',            
            ['ok'], 'ok')

        msg = area.createFormattedText()
        msg.setMinSize(500, 400)
        msg.setText('<tt>' + options + '</tt>')
        dialog.run()        


def setLaunchMessage(text):
    """Launch message is displayed next to the Play button."""
    
    launchText.setText(text)
    launchText.resizeToBestSize()
    ui.getArea(ui.COMMAND).updateLayout()
    
    if text:
        # Clear the message after a short delay.
        launchTextTimer.start(4000)


def startGame(profile):
    """Start the game using the specified profile.

    @param profile A profiles.Profile object.
    """
    setLaunchMessage(language.translate('launch-message'))
    
    options = generateOptions(profile)
    if options == None:
        return

    # Locate the paths and binaries.  Most of these are configured in
    # the .conf files.
    engineBin = st.getSystemString('doomsday-binary')        
    userPath = paths.getUserPath(paths.RUNTIME)

    options += ' -userdir ' + paths.quote(userPath)

    # Put the response file in the user's runtime directory.
    responseFile = os.path.join(userPath, 'Options.rsp')

    file(responseFile, 'w').write(options + "\n")

    # Execute the command line.
    if host.isWindows():
        spawnFunc = os.spawnv
    else:
        spawnFunc = os.spawnvp

    if host.isWindows():
        spawnFunc(os.P_NOWAIT, engineBin, [engineBin, '@' + paths.quote(responseFile)])
    else:
        spawnFunc(os.P_NOWAIT, engineBin, [engineBin, '@' + responseFile])

    # Shut down if the configuration settings say so.
    value = profile.getValue('quit-on-launch')
    if not value or value.getValue() == 'yes':
        # Quit Snowberry.
        events.sendAfter(events.Command('quit'))


def generateOptions(profile):
    """Generate a text string of all the command line options used
    when launching a game.

    @param profile A profiles.Profile object.  The values of settings
    are retrieved from here.

    @return All the options in a single string.  Returns None if there
    is an unresolved addon conflict.
    """
    clearLog()
    
    # Determine which addons are in use.  The final list of addons
    # includes all the addons that must be loaded at launch (defaults;
    # required parts of boxes).
    usedAddonIds = profile.getFinalAddons()
    usedAddons = map(lambda id: ao.get(id), usedAddonIds)

    # Form the command line.
    cmdLine = []
    
    # Determine the settings that apply to the components and
    # addons.
    effectiveSettings = st.getCompatibleSettings(profile)

    # Are there any system-specific options defined?
    if st.isDefined('common-options'):
        cmdLine.append(ex.evaluate(st.getSystemString('common-options'), None))

    # All profiles use the same runtime directory.
    if st.isDefined('doomsday-base'):
        basePath = os.path.abspath(st.getSystemString('doomsday-base'))
        cmdLine.append('-basedir ' + paths.quote(basePath))

    # Determine the components used by the profile.
    for compId in profile.getComponents():
        # Append the component's options to the command line as-is.
        cmdLine.append( st.getComponent(compId).getOption() )

    # Resolve conflicting addons by letting the user choose between
    # them.
    if not resolveAddonConflicts(usedAddons, profile):
        # Launch aborted!
        return None

    # Get the command line options for loading all the addons.
    for addon in usedAddons:
        params = addon.getCommandLine(profile).strip()
        if params:
            cmdLine.append(params)

    # Update IDs of used addons.
    usedAddonsId = map(lambda a: a.getId(), usedAddons)

    # Get the command line options from each setting.
    for setting in effectiveSettings:
        # If the setting's required addons are not being loaded, skip
        # this setting.
        skipThis = False
        for req in setting.getRequiredAddons():
            if req not in usedAddonIds:
                skipThis = True
                break

        if skipThis:
            continue
        
        # All settings can't be used at all times.  In
        # case of a conflict, some of the settings are ignored.
        if setting.isEnabled(profile):
            params = setting.getCommandLine(profile).strip()
            if params:
                cmdLine.append(params)

    # Send a launch options notification.  This is done so that
    # plugins are able to observe/modify the list of launch options.
    events.send(events.LaunchNotify(cmdLine))

    return string.join(cmdLine, ' ')


def clearLog():
    try:
        os.remove(logFileName)
    except:
        pass
        
    global logFile
    logFile = file(logFileName, 'w')

    logFile.write(st.getSystemString('snowberry-title') + ' ' +
                  st.getSystemString('snowberry-version') + "\n")
    logFile.write("Launching on " +
                  time.strftime('%a, %d %b %Y %H:%M:%S') + "\n")
    

def log(problem):
    """Print a message about conflict resolution in the log."""
    
    global logFile

    problemName = problem[0]

    logFile.write("\n%s:\n" % problemName)

    if problemName == 'override':
        source = problem[1]
        targets = problem[2]
        logFile.write("  preferred addon = " + source.getId() + "\n")
        logFile.write("  not loaded = [\n")
        for target, key in targets:
            logFile.write("    %s (because of %s)\n" % (target.getId(), key))
        logFile.write("  ]\n")

    elif problemName == 'provide-conflict':
        logFile.write("  conflicting providers = [\n")
        for source, target, key in problem[1]:
            logFile.write("    %s / %s (because of %s)\n" % (
                source.getId(), target.getId(), key))
        logFile.write("  ]\n")

    elif problemName == 'missing-requirements':
        logFile.write("  %s requires the following = [\n" % problem[1].getId())
        for key in problem[2]:
            logFile.write("    %s\n" % key)
        logFile.write("  ]\n")
        
    elif problemName == 'exclusion-by-category':
        logFile.write("  addon = " + problem[1].getId() + "\n")
        logFile.write("  excluded addons = [\n")
        for a in problem[2]:
            logFile.write("    %s\n" % a.getId())
        logFile.write("  ]\n")

    elif problemName == 'exclusion-by-value':
        logFile.write("  addon = " + problem[1].getId() + "\n")
        logFile.write("  conflicting values = [\n")
        for v in problem[2]:
            logFile.write("    %s\n" % v)
        logFile.write("  ]\n")

    elif problemName == 'exclusion-by-keyword':
        logFile.write("  addon = " + problem[1].getId() + "\n")
        logFile.write("  conflicted by = [\n")
        for a, key in problem[2]:
            logFile.write("    %s (because of %s)\n" % (a.getId(), key))
        logFile.write("  ]\n")

    logFile.flush()
    

def getNextProblem(addons, profile):
    """Gets a list of problems in the set of addons.  Automatically
    resolves overrides.

    @return An array of conflict descriptions.
    """

    problems = []

    # Resolve overrides immediately.  They don't cause any changes to
    # the profile.
    for prob in ao.findConflicts(addons, profile):
        if prob[0] == 'override':
            log(prob)

            # Remove the overrided addons from the set.
            for a, key in prob[2]:
                if a in addons:
                    addons.remove(a)
        else:
            problems.append(prob)

    if len(problems) > 0:
        return problems[0]
    else:
        return None


def resolveAddonConflicts(addons, profile):
    """Resolve all conflicts in the set of the addons.

    @param addons An array of Addon objects.  This array is modified
    in this function.

    @param profile A profiles.Profile object.

    @return Possible return values:
    - True: the conflicts were resolved
    - False: the launch should be cancelled
    """
    global logFile
    global resolving

    # Print the load order into the log.
    logFile.write("Addons to load:\n")
    for a in addons:
        logFile.write("  " + a.getId() + "\n")

    # The resolving process uses a high-first priority.
    addons.reverse()

    #print "CONFLICTS:"
    #print ao.resolveConflicts(addons, profile)

    problem = getNextProblem(addons, profile)

    # We'll loop here until no more problems are found.
    if problem:
        # There are problems to resolve.  Let's open the Conflict
        # Wizard, which will take care of the issue.
        resolving = ResolveStatus()
        resolving.profile = profile
        resolving.dialog, resolving.area = sb.util.dialog.createButtonDialog(
            'conflict-wizard',
            ['cancel', 'continue'], 'continue')

        # Initialize the wizard with the first problem.
        createResolver(problem, addons)

        if resolving.dialog.run() == 'cancel':
            return False

        # We're done.
        resolving = None

    # Restore the original order so that Doomsday gets the highest
    # priority last.
    addons.reverse()

    logFile.write("\nResolved addons:\n")
    for a in addons:
        logFile.write("  " + a.getId() + "\n")
    logFile.flush()
    return True


def createResolver(problem, addons):
    """Create the necessary widgets in the area to resolve the problem.

    @param area An ui.Area object.  Will be cleared.

    @param problem The problem description as returned by
    addons.findConflicts.

    @param addons Set of Addon objects.  This should be modified
    according to the user's selections.
    """
    global resolving

    # Set the general parameters of the resolver.
    resolving.problem = problem
    resolving.addons = addons
    
    resolving.area.clear()
    resolving.area.setWeight(2)

    message = resolving.area.createFormattedText()
    message.setMinSize(250, 250)

    resolving.area.setWeight(1)

    # Enter this problem into the log.
    log(problem)
    
    # The name of the problem.
    problemName = problem[0]

    if problemName == 'exclusion-by-category':
        # The list of excluded addons. 
        names = map(lambda a: language.translate(a.getId()), problem[2])
        names.sort()

        msg = "<ul><li>" + string.join(names, '<li>') + "</ul>"
    
        message.setText(language.expand(
            language.translate('category-conflict-message'),
            language.translate(problem[1].getId()), msg))

        # Create the radio buttons.
        resolving.createConflictButtons(language.translate(problem[1].getId()))

    elif problemName == 'exclusion-by-keyword':
        msg = '<ul>'

        for addon, key in problem[2]:
            msg += '<li>'
            if addon.getId() != key:
                msg += key + " in "
            msg += language.translate(addon.getId())

        msg += '</ul>'

        message.setText(language.expand(
            language.translate('keyword-conflict-message'),
            language.translate(problem[1].getId()), msg))

        # Create the radio buttons.
        resolving.createConflictButtons(language.translate(problem[1].getId()))

    elif problemName == 'exclusion-by-value':
        # The list of excluded values.
        names = []
        for v in problem[2]:
            if language.isDefined(v):
                names.append(language.translate(v))
            else:
                names.append(v)
        msg = "<ul><li>" + string.join(names, '<li>') + "</ul>"
        message.setText(language.expand(
            language.translate('value-conflict-message'),
            language.translate(problem[1].getId()), msg))            

    elif problemName == 'provide-conflict':
        # Collect the conflicting addons into a list.
        resolving.conflicted = []
        keys = []
        for a, b, key in problem[1]:

            if a not in resolving.conflicted:
                resolving.conflicted.append(a)
                
            if b not in resolving.conflicted:
                resolving.conflicted.append(b)
                
            if key not in keys:
                keys.append(key)

        #ao.sortIdentifiersByName(conflicting)

        message.setText(language.expand(
            language.translate('provide-conflict-message'),
            "<ul><li>" + string.join(keys, '<li>') + "</ul>"))

        resolving.list = resolving.area.createList('')
        for a in resolving.conflicted:
            resolving.list.addItem(a.getId())

    elif problemName == 'missing-requirements':

        resolving.conflicted = problem[1]

        res = "<p><ul>"
        res += "<li>" + string.join(problem[2], '<li>') + "</ul>"

        msg = language.expand(
            language.translate('missing-requirements-message'),
            language.translate(problem[1].getId()))
        
        message.setText(msg + res)

        resolving.list = None
        
    resolving.area.updateLayout()


def continueResolving():
    """The user pressed the Continue button in the resolver."""

    # Assume that the currently active profile is the one used for
    # launching.
    profile = resolving.profile

    if resolving.problem[0] == 'provide-conflict':
        # Remove all but the selected addon.
        sel = resolving.list.getSelectedItem()
        if not len(sel):
            return

        # Something was actually selected.  Remove the other
        # addons form the profile.
        for addon in resolving.conflicted:
            if addon.getId() != sel:
                resolving.dontUse(addon)

        problemSolved()

    elif resolving.problem[0] == 'missing-requirements':
        # Remove the addon from the profile.
        addon = resolving.conflicted
        
        profile.dontUseAddon(addon.getId())
        
        if addon in resolving.addons:
            resolving.addons.remove(addon)

        problemSolved()
        
    elif resolving.problem[0] == 'exclusion-by-category':
        if resolving.choices[0].isChecked():
            # Remove the excluded addons.
            for addon in resolving.problem[2]:
                resolving.dontUse(addon)
        else:
            # Remove the addon causing the conflict.
            resolving.dontUse(resolving.problem[1])
        
        problemSolved()

    elif resolving.problem[0] == 'exclusion-by-value':
        resolving.dontUse(resolving.problem[1])
        problemSolved()

    elif resolving.problem[0] == 'exclusion-by-keyword':
        if resolving.choices[0].isChecked():
            # Remove the excluded addons.
            for addon, key in resolving.problem[2]:
                resolving.dontUse(addon)
        else:
            # Remove the addon causing the conflict.
            resolving.dontUse(resolving.problem[1])
        
        problemSolved()


def problemSolved():
    """Called when the current problem has been solved in the wizard."""

    global resolving
    
    problem = getNextProblem(resolving.addons, resolving.profile)

    if problem:
        createResolver(problem, resolving.addons)

    else:
        # We're done.  Close the resolver wizard.
        resolving.dialog.close('ok')
        resolving = None



# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2005
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file tab40_settings.py Settings Tab

import ui, events
import sb.confdb as st
import sb.profdb as pr
import sb.aodb as ao
import widgets as wg
import sb.widget.tab

# Identifiers:
SETTINGS = 'tab-settings'
CATEGORY_AREA = 'settab'

shownSettings = []


def init():
    """Create the Settings page.  No setting widgets are created until
    profiles have been loaded.
    """
    # The setting categories tab area.
    global categoryArea

    area = ui.createTab(SETTINGS)

    # Create the area where all the setting categories are placed.
    categoryArea = area.createTabArea(CATEGORY_AREA,
                                      sb.widget.tab.TabArea.STYLE_FORMATTED)

    createWidgets()

    # Register a notification listener.
    events.addNotifyListener(handleNotify, ['init-done',
                                            'value-changed',
                                            'active-profile-changed',
                                            'addon-database-reloaded'])

    # The addon-settings command is handled here.
    events.addCommandListener(handleCommand, ['show-addon-settings',
                                              'show-snowberry-settings'])


def createWidgets():
    """Create all the setting widgets and tabs.  This is done at
    startup and when addons are installed or uninstalled."""

    # Start with a fresh category tab area.
    categoryArea.removeAllTabs()

    # An array that contains all the existing setting groups.
    global categoryTabs
    categoryTabs = []

    # Create all the category icons.
    allSettingGroups = st.getSettingGroups()

    for group in allSettingGroups:
        area = categoryArea.addTab(group)
        categoryTabs.append((group, area))

    # Addon settings all go inside one tab.
    area = categoryArea.addTab('addons-options')
    categoryTabs.append(('addons-options', area))

    # The Addons tab contains another tab area, where we'll have one
    # tab for each configurable addon that will be loaded.
    global addonArea
    addonArea = area.createTabArea('addontab', sb.widget.tab.TabArea.STYLE_DROP_LIST)

    # Get all addons that exist, get their identifiers, and sort them
    # by name.
    allAddons = map(lambda a: a.getId(), ao.getAddons())
    ao.sortIdentifiersByName(allAddons)

    for addon in allAddons:
        # Only create tabs for the addons that actually have settings
        # defined for them.
        if st.haveSettingsForAddon(addon):
            area = addonArea.addTab(addon)
            categoryTabs.append((addon, area))

            # Each addon may have multiple pages inside its area.
            # These are determined by the groups.
            addonGroups = st.getSettingGroups(addon)

            if len(addonGroups) > 0:
                # Create tabs for each group.
                area = area.createTabArea(addon + '-tab',
                                          sb.widget.tab.TabArea.STYLE_HORIZ_ICON)
                for group in addonGroups:
                    groupId = addon + '-' + group
                    categoryTabs.append((groupId, area.addTab(groupId)))

    # Populate all the tabs with the setting widgets.  When a profile
    # becomes active, the settings that are not applicable will be
    # disabled.  Certain tab pages may be hidden if they are not
    # applicable.
    populateTabs()

    # Request others to create additional widgets in the tabs.
    requestPopulation()

    # Redo the layout.
    ui.getArea(SETTINGS).updateLayout()


def handleCommand(event):
    """Handle commands."""
    if event.hasId('show-addon-settings'):
        # Switch to the Settings tab.
        ui.selectTab(SETTINGS)

        # Switch to the Addons category.
        categoryArea.selectTab('addons-options')

        # Switch to the correct tab.
        addonArea.selectTab(event.addon)
        
    elif event.hasId('show-snowberry-settings'):
        # Change to the Defaults profile.
        pr.setActive(pr.getDefaults())
    
        # Switch to the Settings tab.
        ui.selectTab(SETTINGS)
        
        # Switch to the Snowberry category.
        categoryArea.selectTab('general-options')        


def requestPopulation():
    # Request others to create widgets in the setting tabs, if
    # necessary.
    for areaId, area in categoryTabs:
        events.send(events.PopulateNotify(areaId, area))
        area.updateLayout()


def handleNotify(event):
    """This is called when someone sends a notification event."""

    #if event.hasId('init-done'):

    if (event.hasId('value-changed') or
          event.hasId('active-profile-changed')):
        # Update any settings with value dependencies.
        enableByRequirements(pr.getActive())
        showCompatibleAddons(pr.getActive())

        # Only the defaults profile has the General category.
        if pr.getActive() is pr.getDefaults():
            wasIt = (categoryArea.getSelectedTab() == 'game-options')
            
            categoryArea.showTab('general-options')
            categoryArea.showTab('game-options', False)

            if wasIt:
                categoryArea.selectTab('general-options')
        else:
            wasIt = (categoryArea.getSelectedTab() == 'general-options')

            categoryArea.showTab('game-options')
            categoryArea.showTab('general-options', False)

            if wasIt:
                categoryArea.selectTab('game-options')

            categoryArea.updateIcon('game-options')

    elif event.hasId('addon-database-reloaded'):
        # Since the new addon will most likely introduce new settings,
        # we'll need to recreate the all the settings tabs.
        createWidgets()

        # The newly-created widgets don't have any data in them.  Make
        # them refresh their values.
        pr.refresh()


def populateWithSettings(areaId, area, settings):
    """Create widgets for the given settings in the specified area.

    @param area The area into which the setting widgets will be
    created.

    @param settings An array of Setting objects.

    @param profile A profiles.Profile object, where the values of the
    settings are taken from.
    """
    global shownSettings

    # If no settings were given, we don't need to do anything.
    if len(settings) == 0:
        return

    # Determine if any subgroups need to be formed.
    grouping = {}

    for s in settings:
        name = s.getSubGroup()
        if not name:
            name = ''
        try:
            grouping[name].append(s)
        except KeyError:
            # name wasn't in the array yet.
            grouping[name] = [s]

    orderedGroupKeys = grouping.keys()

    # Sort the subgroups based on the earliest setting order
    # number in each group.
    def getEarliestOrder(settings):
        earliest = 0
        for s in settings:
            if earliest == 0 or s.getOrder() < earliest:
                earliest = s.getOrder()
        return earliest

    orderedGroupKeys.sort(lambda a,b: cmp(getEarliestOrder(grouping[a]),
                                          getEarliestOrder(grouping[b])))

    # Create widgets for all the settings.
    for i in orderedGroupKeys:
        subGroupSettings = grouping[i]

        # Sort the settings based on the order.
        subGroupSettings.sort(lambda a, b: cmp(a.getOrder(), b.getOrder()))

        if i == '':
            # No grouping.
            where = area
        else:
            # Create a subgroup.
            where = area.createArea(border=2, boxedWithTitle=i)
            where.setWeight(0)

        for s in subGroupSettings:
            settingArea, settingWidget = where.createSetting(s)
            if settingArea:
                # Add it into the array of shown settings.
                shownSettings.append((s, settingArea))

    area.updateLayout()


def populateTabs():
    """Create setting widgets in all the tabs."""

    global shownSettings
    shownSettings = []

    for areaId, area in categoryTabs:
        # Use the appropriate layout parameters.
        area.setWeight(0)
        #area.setBorder(2)

        # Create a title inside the tab.
        #if areaId != 'addons-options' and areaId[-8:] == '-options':
        #    area.createText(areaId).setHeadingStyle()
        #    area.createLine()

        def settingFilter(s):
            if len(s.getRequiredAddons()) > 0:
                ident = s.getRequiredAddons()[0]
                properTab = ident

                # The addon does not exist?
                if not ao.exists(ident) or ao.get(ident).isUninstalled():
                    return False

                if s.getGroup():
                    # Addon setting with a group.
                    properTab += '-' + s.getGroup()
            elif not s.getGroup():
                properTab = 'general-options'
            else:
                properTab = s.getGroup()
            return properTab == areaId

        # Collect a list of all the settings for this tab (group).
        tabSettings = st.getSettings(settingFilter)

        populateWithSettings(areaId, area, tabSettings)


def enableByRequirements(profile):
    """Disable settings that can't be used at the moment.  The
    requirements are resolved based on the values in the specified
    profile.

    @param profile A profiles.Profile object.
    """
    for setting, area in shownSettings:
        area.enable(setting.isEnabled(profile))


def showCompatibleAddons(profile):
    """Show only the addons (tabs) that are compatible with the
    current profile.

    @param profile Profile to check against.
    """
    compatibleAddonIds = map(lambda a: a.getId(),
                              ao.getAvailableAddons(profile))

    for identifier in addonArea.getTabs():
        addonArea.showTab(identifier, identifier in compatibleAddonIds)



import os
import sys

def get_arg(label):
    """Find the value for the command line option @a label."""
    if label in sys.argv:
        return sys.argv[sys.argv.index(label) + 1]
    return None

BUILD_AUTHOR_NAME = "skyjake"
BUILD_AUTHOR_EMAIL = "skyjake@users.sourceforge.net"
BUILD_URI = "http://code.iki.fi/builds"
RFC_TIME = "%a, %d %b %Y %H:%M:%S +0000"
if 'HOME' in os.environ:
    EVENT_DIR = os.path.join(os.environ['HOME'], 'BuildMaster')
else:
    EVENT_DIR = '.'
DISTRIB_DIR = '.'
APT_REPO_DIR = ''
TAG_MODIFIER = ''
BRANCH = 'master'

val = get_arg('--distrib')
if val is not None: DISTRIB_DIR = val

val = get_arg('--events')
if val is not None: EVENT_DIR = val

val = get_arg('--apt')
if val is not None: APT_REPO_DIR = val

val = get_arg('--branch')
if val is not None: BRANCH = val

val = get_arg('--tagmod')
if val is not None: TAG_MODIFIER = val

# Determine APT repository path.
oldCwd = os.getcwd()
if DISTRIB_DIR:
    DISTRIB_DIR = os.path.abspath(DISTRIB_DIR)
    os.chdir(DISTRIB_DIR)
import build_version
build_version.find_version(quiet=True)
if build_version.DOOMSDAY_RELEASE_TYPE == 'Stable':
    APT_DIST = 'dists/stable'
    APT_CONF_FILE = '~/Dropbox/APT/ftparchive-release-stable.conf'
else:
    APT_DIST = 'dists/unstable'
    APT_CONF_FILE = '~/Dropbox/APT/ftparchive-release.conf'
os.chdir(oldCwd)
